from typing import List

from pydantic import Field

from beanie import Document, Indexed, Link
from beanie.odm.fields import BackLink
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2


class WindowAPI(Document):
    x: int
    y: int


class DoorAPI(Document):
    t: int = 10


class RoofAPI(Document):
    r: int = 100


class HouseAPI(Document):
    windows: List[Link[WindowAPI]]
    name: Indexed(str)
    height: Indexed(int) = 2


class House(Document):
    name: str
    owner: Link["Person"]


class Person(Document):
    name: str
    house: BackLink[House] = (
        Field(json_schema_extra={"original_field": "owner"})
        if IS_PYDANTIC_V2
        else Field(original_field="owner")
    )



from typing import Optional, Union

import pytest
from pydantic import BaseModel
from typing_extensions import Annotated

from beanie import Document, Link
from beanie.odm.fields import Indexed
from beanie.odm.utils.pydantic import get_model_fields
from beanie.odm.utils.typing import extract_id_class, get_index_attributes


class Lock(Document):
    k: int


class TestTyping:
    def test_extract_id_class(self):
        # Union
        assert extract_id_class(Union[str, int]) is str
        assert extract_id_class(Union[str, None]) is str
        assert extract_id_class(Union[str, None, int]) is str
        # Optional
        assert extract_id_class(Optional[str]) is str
        # Link
        assert extract_id_class(Link[Lock]) == Lock

    @pytest.mark.parametrize(
        "type,result",
        (
            (str, None),
            (Indexed(str), (1, {})),
            (Indexed(str, "text", unique=True), ("text", {"unique": True})),
            (Annotated[str, Indexed()], (1, {})),
            (
                Annotated[str, "other metadata", Indexed(unique=True)],
                (1, {"unique": True}),
            ),
            (Annotated[str, "other metadata"], None),
        ),
    )
    def test_get_index_attributes(self, type, result):
        class Foo(BaseModel):
            bar: type

        field = get_model_fields(Foo)["bar"]
        assert get_index_attributes(field) == result



from contextlib import asynccontextmanager

from fastapi import FastAPI
from pymongo import AsyncMongoClient

from beanie import init_beanie
from tests.conftest import Settings
from tests.fastapi.models import (
    DoorAPI,
    House,
    HouseAPI,
    Person,
    RoofAPI,
    WindowAPI,
)
from tests.fastapi.routes import house_router


@asynccontextmanager
async def live_span(_: FastAPI):
    # CREATE ASYNC PYMONGO CLIENT
    client = AsyncMongoClient(Settings().mongodb_dsn)

    # INIT BEANIE
    await init_beanie(
        client.beanie_db,
        document_models=[House, Person, HouseAPI, WindowAPI, DoorAPI, RoofAPI],
    )

    yield


app = FastAPI(lifespan=live_span)

# ADD ROUTES
app.include_router(house_router, prefix="/v1", tags=["house"])



import asyncio

from pymongo import AsyncMongoClient

from beanie import Document, init_beanie


class SampleModel(Document):
    s: str = "TEST"
    i: int = 10


class SampleModel2(SampleModel): ...


class SampleModel3(SampleModel2): ...


class TestConcurrency:
    async def test_without_init(self, settings):
        clients = []
        for _ in range(10):
            client = AsyncMongoClient(settings.mongodb_dsn)
            clients.append(client)
            db = client[settings.mongodb_db_name]
            await init_beanie(
                db, document_models=[SampleModel3, SampleModel, SampleModel2]
            )

            async def insert_find():
                await SampleModel2().insert()
                docs = await SampleModel2.find(SampleModel2.i == 10).to_list()
                return docs

            await asyncio.gather(*[insert_find() for _ in range(10)])

        await SampleModel2.delete_all()

        [await client.close() for client in clients]



from beanie.odm.operators.find.bitwise import (
    BitsAllClear,
    BitsAllSet,
    BitsAnyClear,
    BitsAnySet,
)
from tests.odm.models import Sample


async def test_bits_all_clear():
    q = BitsAllClear(Sample.integer, "smth")
    assert q == {"integer": {"$bitsAllClear": "smth"}}


async def test_bits_all_set():
    q = BitsAllSet(Sample.integer, "smth")
    assert q == {"integer": {"$bitsAllSet": "smth"}}


async def test_any_clear():
    q = BitsAnyClear(Sample.integer, "smth")
    assert q == {"integer": {"$bitsAnyClear": "smth"}}


async def test_any_set():
    q = BitsAnySet(Sample.integer, "smth")
    assert q == {"integer": {"$bitsAnySet": "smth"}}



import pytest
from pydantic import Field
from pydantic.main import BaseModel
from pymongo.errors import OperationFailure

from beanie.odm.enums import SortDirection
from beanie.odm.utils.find import construct_lookup_queries
from tests.odm.models import DocumentWithTextIndexAndLink, Sample


async def test_aggregate(preset_documents):
    q = Sample.aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}]
    )
    assert q.get_aggregation_pipeline() == [
        {"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}
    ]
    result = await q.to_list()
    assert len(result) == 4
    assert {"_id": "test_3", "total": 3} in result
    assert {"_id": "test_1", "total": 3} in result
    assert {"_id": "test_0", "total": 0} in result
    assert {"_id": "test_2", "total": 6} in result


async def test_aggregate_with_filter(preset_documents):
    q = Sample.find(Sample.increment >= 4).aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}]
    )
    assert q.get_aggregation_pipeline() == [
        {"$match": {"increment": {"$gte": 4}}},
        {"$group": {"_id": "$string", "total": {"$sum": "$integer"}}},
    ]
    result = await q.to_list()
    assert len(result) == 3
    assert {"_id": "test_1", "total": 2} in result
    assert {"_id": "test_2", "total": 6} in result
    assert {"_id": "test_3", "total": 3} in result


async def test_aggregate_with_sort_skip(preset_documents):
    q = Sample.find(sort="_id", skip=2).aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}]
    )
    assert q.get_aggregation_pipeline() == [
        {"$group": {"_id": "$string", "total": {"$sum": "$integer"}}},
        {"$sort": {"_id": SortDirection.ASCENDING}},
        {"$skip": 2},
    ]
    assert await q.to_list() == [
        {"_id": "test_2", "total": 6},
        {"_id": "test_3", "total": 3},
    ]


async def test_aggregate_with_sort_limit(preset_documents):
    q = Sample.find(sort="_id", limit=2).aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}]
    )
    assert q.get_aggregation_pipeline() == [
        {"$group": {"_id": "$string", "total": {"$sum": "$integer"}}},
        {"$sort": {"_id": SortDirection.ASCENDING}},
        {"$limit": 2},
    ]
    assert await q.to_list() == [
        {"_id": "test_0", "total": 0},
        {"_id": "test_1", "total": 3},
    ]


async def test_aggregate_with_projection_model(preset_documents):
    class OutputItem(BaseModel):
        id: str = Field(None, alias="_id")
        total: int

    ids = []
    q = Sample.find(Sample.increment >= 4).aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}],
        projection_model=OutputItem,
    )
    assert q.get_aggregation_pipeline() == [
        {"$match": {"increment": {"$gte": 4}}},
        {"$group": {"_id": "$string", "total": {"$sum": "$integer"}}},
        {"$project": {"_id": 1, "total": 1}},
    ]
    async for i in q:
        if i.id == "test_1":
            assert i.total == 2
        elif i.id == "test_2":
            assert i.total == 6
        elif i.id == "test_3":
            assert i.total == 3
        else:
            raise KeyError
        ids.append(i.id)
    assert set(ids) == {"test_1", "test_2", "test_3"}


async def test_aggregate_with_session(preset_documents, session):
    q = Sample.find(Sample.increment >= 4).aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}],
        session=session,
    )
    assert q.session == session

    q = Sample.find(Sample.increment >= 4, session=session).aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}]
    )
    assert q.session == session

    result = await q.to_list()

    assert len(result) == 3
    assert {"_id": "test_1", "total": 2} in result
    assert {"_id": "test_2", "total": 6} in result
    assert {"_id": "test_3", "total": 3} in result


async def test_aggregate_pymongo_kwargs(preset_documents):
    with pytest.raises(OperationFailure):
        await (
            Sample.find(Sample.increment >= 4)
            .aggregate(
                [
                    {
                        "$group": {
                            "_id": "$string",
                            "total": {"$sum": "$integer"},
                        }
                    }
                ],
                wrong=True,
            )
            .to_list()
        )


async def test_clone(preset_documents):
    q = Sample.find(Sample.increment >= 4).aggregate(
        [{"$group": {"_id": "$string", "total": {"$sum": "$integer"}}}]
    )
    new_q = q.clone()
    new_q.aggregation_pipeline.append({"a": "b"})
    assert q.get_aggregation_pipeline() == [
        {"$match": {"increment": {"$gte": 4}}},
        {"$group": {"_id": "$string", "total": {"$sum": "$integer"}}},
    ]
    assert new_q.get_aggregation_pipeline() == [
        {"$match": {"increment": {"$gte": 4}}},
        {"$group": {"_id": "$string", "total": {"$sum": "$integer"}}},
        {"a": "b"},
    ]


@pytest.mark.parametrize("text_query_count", [0, 1, 2])
@pytest.mark.parametrize("non_text_query_count", [0, 1, 2])
async def test_with_text_queries(
    text_query_count: int, non_text_query_count: int
):
    text_query = {"$text": {"$search": "text_search"}}
    non_text_query = {"s": "test_string"}
    aggregation_pipeline = [{"$count": "count"}]
    queries = []

    if text_query_count:
        queries.append(text_query)
        if text_query_count > 1:
            queries.append(text_query)

    if non_text_query_count:
        queries.append(non_text_query)
        if non_text_query_count > 1:
            queries.append(non_text_query)

    query = DocumentWithTextIndexAndLink.find(*queries, fetch_links=True)

    expected_aggregation_pipeline = []
    if text_query_count:
        expected_aggregation_pipeline.append(
            {"$match": text_query}
            if text_query_count == 1
            else {"$match": {"$and": [text_query, text_query]}}
        )

    expected_aggregation_pipeline.extend(
        construct_lookup_queries(query.document_model)
    )

    if non_text_query_count:
        expected_aggregation_pipeline.append(
            {"$match": non_text_query}
            if non_text_query_count == 1
            else {"$match": {"$and": [non_text_query, non_text_query]}}
        )

    expected_aggregation_pipeline.extend(aggregation_pipeline)

    assert (
        query.build_aggregation_pipeline(*aggregation_pipeline)
        == expected_aggregation_pipeline
    )



from typing import TYPE_CHECKING, Any, Dict, Generator, Mapping, Optional, Type

from pymongo import DeleteMany as DeleteManyPyMongo
from pymongo import DeleteOne as DeleteOnePyMongo
from pymongo.asynchronous.client_session import AsyncClientSession
from pymongo.results import DeleteResult

from beanie.odm.bulk import BulkWriter
from beanie.odm.interfaces.clone import CloneInterface
from beanie.odm.interfaces.session import SessionMethods

if TYPE_CHECKING:
    from beanie.odm.documents import DocType


class DeleteQuery(SessionMethods, CloneInterface):
    """
    Deletion Query
    """

    def __init__(
        self,
        document_model: Type["DocType"],
        find_query: Mapping[str, Any],
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs: Any,
    ):
        self.document_model = document_model
        self.find_query = find_query
        self.session: Optional[AsyncClientSession] = None
        self.bulk_writer = bulk_writer
        self.pymongo_kwargs: Dict[str, Any] = pymongo_kwargs


class DeleteMany(DeleteQuery):
    def __await__(
        self,
    ) -> Generator[DeleteResult, None, Optional[DeleteResult]]:
        """
        Run the query
        :return:
        """
        if self.bulk_writer is None:
            return (
                yield from self.document_model.get_pymongo_collection()
                .delete_many(
                    self.find_query,
                    session=self.session,
                    **self.pymongo_kwargs,
                )
                .__await__()
            )
        else:
            self.bulk_writer.add_operation(
                self.document_model,
                DeleteManyPyMongo(self.find_query, **self.pymongo_kwargs),
            )
            return None


class DeleteOne(DeleteQuery):
    def __await__(
        self,
    ) -> Generator[DeleteResult, None, Optional[DeleteResult]]:
        """
        Run the query
        :return:
        """
        if self.bulk_writer is None:
            return (
                yield from self.document_model.get_pymongo_collection()
                .delete_one(
                    self.find_query,
                    session=self.session,
                    **self.pymongo_kwargs,
                )
                .__await__()
            )
        else:
            self.bulk_writer.add_operation(
                self.document_model,
                DeleteOnePyMongo(self.find_query),
                **self.pymongo_kwargs,
            )
            return None



from beanie.odm.utils.pydantic import IS_PYDANTIC_V2

if IS_PYDANTIC_V2:
    from beanie.odm.custom_types.decimal import DecimalAnnotation
else:
    from decimal import Decimal as DecimalAnnotation

__all__ = [
    "DecimalAnnotation",
]



import pytest

from tests.odm.models import (
    DocumentWithExtras,
    DocumentWithExtrasKw,
    DocumentWithPydanticConfig,
)


async def test_pydantic_extras():
    doc = DocumentWithExtras(num_1=2)
    doc.extra_value = "foo"
    await doc.save()

    loaded_doc = await DocumentWithExtras.get(doc.id)

    assert loaded_doc.extra_value == "foo"


@pytest.mark.skip(reason="setting extra to allow via class kwargs not working")
async def test_pydantic_extras_kw():
    doc = DocumentWithExtrasKw(num_1=2)
    doc.extra_value = "foo"
    await doc.save()

    loaded_doc = await DocumentWithExtras.get(doc.id)

    assert loaded_doc.extra_value == "foo"


async def test_fail_with_no_extras():
    doc = DocumentWithPydanticConfig(num_1=2)
    with pytest.raises(ValueError):
        doc.extra_value = "foo"



from beanie.odm.operators.find.geospatial import (
    Box,
    GeoIntersects,
    GeoWithin,
    Near,
    NearSphere,
)
from tests.odm.models import Sample


async def test_geo_intersects():
    q = GeoIntersects(
        Sample.geo, geo_type="Polygon", coordinates=[[1, 1], [2, 2], [3, 3]]
    )
    assert q == {
        "geo": {
            "$geoIntersects": {
                "$geometry": {
                    "type": "Polygon",
                    "coordinates": [[1, 1], [2, 2], [3, 3]],
                }
            }
        }
    }


async def test_geo_within():
    q = GeoWithin(
        Sample.geo, geo_type="Polygon", coordinates=[[1, 1], [2, 2], [3, 3]]
    )
    assert q == {
        "geo": {
            "$geoWithin": {
                "$geometry": {
                    "type": "Polygon",
                    "coordinates": [[1, 1], [2, 2], [3, 3]],
                }
            }
        }
    }


async def test_box():
    q = Box(Sample.geo, lower_left=[1, 3], upper_right=[2, 4])
    assert q == {"geo": {"$geoWithin": {"$box": [[1, 3], [2, 4]]}}}


async def test_near():
    q = Near(Sample.geo, longitude=1.1, latitude=2.2)
    assert q == {
        "geo": {
            "$near": {
                "$geometry": {"type": "Point", "coordinates": [1.1, 2.2]}
            }
        }
    }

    q = Near(Sample.geo, longitude=1.1, latitude=2.2, max_distance=1)
    assert q == {
        "geo": {
            "$near": {
                "$geometry": {"type": "Point", "coordinates": [1.1, 2.2]},
                "$maxDistance": 1,
            }
        }
    }

    q = Near(
        Sample.geo,
        longitude=1.1,
        latitude=2.2,
        max_distance=1,
        min_distance=0.5,
    )
    assert q == {
        "geo": {
            "$near": {
                "$geometry": {"type": "Point", "coordinates": [1.1, 2.2]},
                "$maxDistance": 1,
                "$minDistance": 0.5,
            }
        }
    }


async def test_near_sphere():
    q = NearSphere(Sample.geo, longitude=1.1, latitude=2.2)
    assert q == {
        "geo": {
            "$nearSphere": {
                "$geometry": {"type": "Point", "coordinates": [1.1, 2.2]}
            }
        }
    }

    q = NearSphere(Sample.geo, longitude=1.1, latitude=2.2, max_distance=1)
    assert q == {
        "geo": {
            "$nearSphere": {
                "$geometry": {"type": "Point", "coordinates": [1.1, 2.2]},
                "$maxDistance": 1,
            }
        }
    }

    q = NearSphere(
        Sample.geo,
        longitude=1.1,
        latitude=2.2,
        max_distance=1,
        min_distance=0.5,
    )
    assert q == {
        "geo": {
            "$nearSphere": {
                "$geometry": {"type": "Point", "coordinates": [1.1, 2.2]},
                "$maxDistance": 1,
                "$minDistance": 0.5,
            }
        }
    }



import pytest
from pymongo import AsyncMongoClient

from beanie.odm.utils.pydantic import IS_PYDANTIC_V2

if IS_PYDANTIC_V2:
    from pydantic_settings import BaseSettings
else:
    from pydantic import BaseSettings


class Settings(BaseSettings):
    mongodb_dsn: str = "mongodb://localhost:27017/beanie_db"
    mongodb_db_name: str = "beanie_db"


@pytest.fixture
def settings():
    return Settings()


@pytest.fixture
async def cli(settings):
    client = AsyncMongoClient(settings.mongodb_dsn)

    yield client

    await client.close()


@pytest.fixture
def db(cli, settings):
    return cli[settings.mongodb_db_name]



from abc import abstractmethod
from typing import Any, Dict, Optional, Type, TypeVar, Union, overload

from pydantic import BaseModel
from pymongo.asynchronous.client_session import AsyncClientSession

from beanie.odm.fields import ExpressionField
from beanie.odm.queries.aggregation import AggregationQuery
from beanie.odm.queries.find import FindMany

DocType = TypeVar("DocType", bound="AggregateInterface")
DocumentProjectionType = TypeVar("DocumentProjectionType", bound=BaseModel)


class AggregateInterface:
    @classmethod
    @abstractmethod
    def find_all(cls) -> FindMany:
        pass

    @overload
    @classmethod
    def aggregate(
        cls: Type[DocType],
        aggregation_pipeline: list,
        projection_model: None = None,
        session: Optional[AsyncClientSession] = None,
        ignore_cache: bool = False,
        **pymongo_kwargs: Any,
    ) -> AggregationQuery[Dict[str, Any]]: ...

    @overload
    @classmethod
    def aggregate(
        cls: Type[DocType],
        aggregation_pipeline: list,
        projection_model: Type[DocumentProjectionType],
        session: Optional[AsyncClientSession] = None,
        ignore_cache: bool = False,
        **pymongo_kwargs: Any,
    ) -> AggregationQuery[DocumentProjectionType]: ...

    @classmethod
    def aggregate(
        cls: Type[DocType],
        aggregation_pipeline: list,
        projection_model: Optional[Type[DocumentProjectionType]] = None,
        session: Optional[AsyncClientSession] = None,
        ignore_cache: bool = False,
        **pymongo_kwargs: Any,
    ) -> Union[
        AggregationQuery[Dict[str, Any]],
        AggregationQuery[DocumentProjectionType],
    ]:
        """
        Aggregate over collection.
        Returns [AggregationQuery](query.md#aggregationquery) query object
        :param aggregation_pipeline: list - aggregation pipeline
        :param projection_model: Type[BaseModel]
        :param session: Optional[AsyncClientSession] - pymongo session.
        :param ignore_cache: bool
        :param **pymongo_kwargs: pymongo native parameters for aggregate operation
        :return: [AggregationQuery](query.md#aggregationquery)
        """
        return cls.find_all().aggregate(
            aggregation_pipeline=aggregation_pipeline,
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            **pymongo_kwargs,
        )

    @classmethod
    async def sum(
        cls,
        field: Union[ExpressionField, float, int, str],
        session: Optional[AsyncClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[float]:
        """
        Sum of values of the given field over the entire collection.

        Example:

        ```python

        class Sample(Document):
            price: int

        sum_count = await Document.sum(Sample.price)

        ```

        :param field: Union[ExpressionField, float, int, str]
        :param session: Optional[AsyncClientSession] - pymongo session
        :param ignore_cache: bool
        :return: float - sum. None if there are no items.
        """
        return await cls.find_all().sum(field, session, ignore_cache)

    @classmethod
    async def avg(
        cls,
        field: Union[ExpressionField, float, int, str],
        session: Optional[AsyncClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[float]:
        """
        Average of values of the given field over the entire collection.

        Example:

        ```python

        class Sample(Document):
            price: int

        avg_count = await Document.avg(Sample.price)
        ```

        :param field: Union[ExpressionField, float, int, str]
        :param session: Optional[AsyncClientSession] - pymongo session
        :param ignore_cache: bool
        :return: Optional[float] - avg. None if there are no items.
        """
        return await cls.find_all().avg(field, session, ignore_cache)

    @classmethod
    async def max(
        cls,
        field: Union[ExpressionField, str, Any],
        session: Optional[AsyncClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[Any]:
        """
        Max of the values of the given field over the entire collection.

        Example:

        ```python

        class Sample(Document):
            price: int

        max_count = await Document.max(Sample.price)
        ```

        :param field: Union[ExpressionField, str, Any]
        :param session: Optional[AsyncClientSession] - pymongo session
        :return: Any - max value. None if there are no items.
        """
        return await cls.find_all().max(field, session, ignore_cache)

    @classmethod
    async def min(
        cls,
        field: Union[ExpressionField, str, Any],
        session: Optional[AsyncClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[Any]:
        """
        Min of the values of the given field over the entire collection.

        Example:

        ```python

        class Sample(Document):
            price: int

        min_count = await Document.min(Sample.price)
        ```

        :param field: Union[ExpressionField, str, Any]
        :param session: Optional[AsyncClientSession] - pymongo session
        :return: Any - min value. None if there are no items.
        """
        return await cls.find_all().min(field, session, ignore_cache)



from enum import Enum
from typing import Annotated, Any, Dict, Optional

from pydantic import BaseModel, Field


class Granularity(str, Enum):
    """
    Time Series Granuality
    """

    seconds = "seconds"
    minutes = "minutes"
    hours = "hours"


class TimeSeriesConfig(BaseModel):
    """
    Time Series Collection config
    """

    time_field: str
    meta_field: Optional[str] = None
    granularity: Optional[Granularity] = None
    bucket_max_span_seconds: Optional[int] = None
    bucket_rounding_second: Annotated[
        Optional[int],
        Field(
            deprecated="This field is deprecated in favor of "
            "'bucket_rounding_seconds'.",
        ),
    ] = None
    bucket_rounding_seconds: Optional[int] = None
    expire_after_seconds: Optional[int] = None

    def build_query(self, collection_name: str) -> Dict[str, Any]:
        res: Dict[str, Any] = {"name": collection_name}
        timeseries: Dict[str, Any] = {"timeField": self.time_field}
        if self.meta_field is not None:
            timeseries["metaField"] = self.meta_field
        if self.granularity is not None:
            timeseries["granularity"] = self.granularity
        if self.bucket_max_span_seconds is not None:
            timeseries["bucketMaxSpanSeconds"] = self.bucket_max_span_seconds
        if self.bucket_rounding_second is not None:  # Deprecated field
            timeseries["bucketRoundingSeconds"] = self.bucket_rounding_second
        if self.bucket_rounding_seconds is not None:
            timeseries["bucketRoundingSeconds"] = self.bucket_rounding_seconds
        res["timeseries"] = timeseries
        if self.expire_after_seconds is not None:
            res["expireAfterSeconds"] = self.expire_after_seconds
        return res



import pymongo

from beanie.odm.fields import PydanticObjectId
from tests.odm.models import DocumentTestModel


async def test_get(document):
    new_document = await DocumentTestModel.get(document.id)
    assert new_document == document


async def test_get_not_found(document):
    new_document = await DocumentTestModel.get(PydanticObjectId())
    assert new_document is None


async def test_find_one(documents):
    inserted_one = await documents(1, "kipasa")
    await documents(10, "smthe else")
    expected_doc_id = PydanticObjectId(inserted_one[0])
    new_document = await DocumentTestModel.find_one({"test_str": "kipasa"})
    assert new_document.id == expected_doc_id


async def test_find_one_not_found(documents):
    await documents(10, "smthe else")

    new_document = await DocumentTestModel.find_one({"test_str": "wrong"})
    assert new_document is None


async def test_find_one_more_than_one_found(documents):
    await documents(10, "one")
    await documents(10, "two")
    new_document = await DocumentTestModel.find_one({"test_str": "one"})
    assert new_document.test_str == "one"


async def test_find_all(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_all().to_list()
    assert len(result) == 7


async def test_find_all_limit(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_all(limit=5).to_list()
    assert len(result) == 5


async def test_find_all_skip(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_all(skip=1).to_list()
    assert len(result) == 6


async def test_find_all_sort(documents):
    await documents(4, "uno", True)
    await documents(2, "dos", True)
    await documents(1, "cuatro", True)
    result = await DocumentTestModel.find_all(
        sort=[
            ("test_str", pymongo.ASCENDING),
            ("test_int", pymongo.DESCENDING),
        ]
    ).to_list()
    assert result[0].test_str == "cuatro"
    assert result[1].test_str == result[2].test_str == "dos"
    assert (
        result[3].test_str
        == result[4].test_str
        == result[5].test_str
        == result[5].test_str
        == "uno"
    )

    assert result[1].test_int >= result[2].test_int
    assert (
        result[3].test_int
        >= result[4].test_int
        >= result[5].test_int
        >= result[6].test_int
    )


async def test_find_many(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_many(
        DocumentTestModel.test_str == "uno"
    ).to_list()
    assert len(result) == 4


async def test_find_many_limit(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_many(
        {"test_str": "uno"}, limit=2
    ).to_list()
    assert len(result) == 2


async def test_find_many_skip(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_many(
        {"test_str": "uno"}, skip=1
    ).to_list()
    assert len(result) == 3


async def test_find_many_sort(documents):
    await documents(4, "uno", True)
    await documents(2, "dos", True)
    await documents(1, "cuatro", True)
    result = await DocumentTestModel.find_many(
        {"test_str": "uno"}, sort="test_int"
    ).to_list()
    assert (
        result[0].test_int
        <= result[1].test_int
        <= result[2].test_int
        <= result[3].test_int
    )

    result = await DocumentTestModel.find_many(
        {"test_str": "uno"}, sort=[("test_int", pymongo.DESCENDING)]
    ).to_list()
    assert (
        result[0].test_int
        >= result[1].test_int
        >= result[2].test_int
        >= result[3].test_int
    )


async def test_find_many_not_found(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_many({"test_str": "wrong"}).to_list()
    assert len(result) == 0


async def test_get_with_session(document, session):
    new_document = await DocumentTestModel.get(document.id, session=session)
    assert new_document == document


async def test_find_one_with_session(documents, session):
    inserted_one = await documents(1, "kipasa")
    await documents(10, "smthe else")

    expected_doc_id = PydanticObjectId(inserted_one[0])

    new_document = await DocumentTestModel.find_one(
        {"test_str": "kipasa"}, session=session
    )
    assert new_document.id == expected_doc_id


async def test_find_all_with_session(documents, session):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_all(session=session).to_list()
    assert len(result) == 7


async def test_find_many_with_session(documents, session):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.find_many(
        {"test_str": "uno"}, session=session
    ).to_list()
    assert len(result) == 4



import sys
from pathlib import Path

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from beanie import __version__


def parse_version_from_pyproject():
    pyproject = Path(__file__).parent.parent / "pyproject.toml"
    with pyproject.open("rb") as f:
        toml_data = tomllib.load(f)
    return toml_data["project"]["version"]


def test_version():
    assert __version__ == parse_version_from_pyproject()



import asyncio
import logging
import os
import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import Any

import click

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from beanie.migrations import template
from beanie.migrations.database import DBHandler
from beanie.migrations.models import RunningDirections, RunningMode
from beanie.migrations.runner import MigrationNode

logging.basicConfig(format="%(message)s", level=logging.INFO)


class MigrationSettings:
    def __init__(self, **kwargs: Any):
        self.direction = (
            kwargs.get("direction")
            or self.get_env_value("direction")
            or self.get_from_toml("direction")
            or RunningDirections.FORWARD
        )

        self.distance = int(
            kwargs.get("distance")
            or self.get_env_value("distance")
            or self.get_from_toml("distance")
            or 0
        )
        self.connection_uri = str(
            kwargs.get("connection_uri")
            or self.get_env_value("connection_uri")
            or self.get_from_toml("connection_uri")
        )
        self.database_name = str(
            kwargs.get("database_name")
            or self.get_env_value("database_name")
            or self.get_from_toml("database_name")
        )
        self.path = Path(
            kwargs.get("path")
            or self.get_env_value("path")
            or self.get_from_toml("path")
        )
        self.allow_index_dropping = bool(
            kwargs.get("allow_index_dropping")
            or self.get_env_value("allow_index_dropping")
            or self.get_from_toml("allow_index_dropping")
            or False
        )
        self.use_transaction = bool(kwargs.get("use_transaction"))

    @staticmethod
    def get_env_value(field_name) -> Any:
        if field_name == "connection_uri":
            value = (
                os.environ.get("BEANIE_URI")
                or os.environ.get("BEANIE_CONNECTION_URI")
                or os.environ.get("BEANIE_CONNECTION_STRING")
                or os.environ.get("BEANIE_MONGODB_DSN")
                or os.environ.get("BEANIE_MONGODB_URI")
                or os.environ.get("beanie_uri")
                or os.environ.get("beanie_connection_uri")
                or os.environ.get("beanie_connection_string")
                or os.environ.get("beanie_mongodb_dsn")
                or os.environ.get("beanie_mongodb_uri")
            )
        elif field_name == "database_name":
            value = (
                os.environ.get("BEANIE_DB")
                or os.environ.get("BEANIE_DB_NAME")
                or os.environ.get("BEANIE_DATABASE_NAME")
                or os.environ.get("beanie_db")
                or os.environ.get("beanie_db_name")
                or os.environ.get("beanie_database_name")
            )
        else:
            value = os.environ.get(
                f"BEANIE_{field_name.upper()}"
            ) or os.environ.get(f"beanie_{field_name.lower()}")
        return value

    @staticmethod
    def get_from_toml(field_name) -> Any:
        path = Path("pyproject.toml")
        if path.is_file():
            with path.open("rb") as f:
                toml_data = tomllib.load(f)
            val = (
                toml_data.get("tool", {})
                .get("beanie", {})
                .get("migrations", {})
            )
        else:
            val = {}
        return val.get(field_name)


@click.group()
def migrations():
    pass


async def run_migrate(settings: MigrationSettings):
    DBHandler.set_db(settings.connection_uri, settings.database_name)
    root = await MigrationNode.build(settings.path)
    mode = RunningMode(
        direction=settings.direction, distance=settings.distance
    )
    await root.run(
        mode=mode,
        allow_index_dropping=settings.allow_index_dropping,
        use_transaction=settings.use_transaction,
    )

    # Cleanup
    client = DBHandler.get_cli()
    if client:
        await client.close()


@migrations.command()
@click.option(
    "--forward",
    "direction",
    required=False,
    flag_value="FORWARD",
    help="Roll the migrations forward. This is default",
)
@click.option(
    "--backward",
    "direction",
    required=False,
    flag_value="BACKWARD",
    help="Roll the migrations backward",
)
@click.option(
    "-d",
    "--distance",
    required=False,
    help="How many migrations should be done since the current? "
    "0 - all the migrations. Default is 0",
)
@click.option(
    "-uri",
    "--connection-uri",
    required=False,
    type=str,
    help="MongoDB connection URI",
)
@click.option(
    "-db", "--database_name", required=False, type=str, help="DataBase name"
)
@click.option(
    "-p",
    "--path",
    required=False,
    type=str,
    help="Path to the migrations directory",
)
@click.option(
    "--allow-index-dropping/--forbid-index-dropping",
    required=False,
    default=False,
    help="if allow-index-dropping is set, Beanie will drop indexes from your collection",
)
@click.option(
    "--use-transaction/--no-use-transaction",
    required=False,
    default=True,
    help="Enable or disable the use of transactions during migration. "
    "When enabled (--use-transaction), Beanie uses transactions for migration, "
    "which necessitates a replica set. When disabled (--no-use-transaction), "
    "migrations occur without transactions.",
)
def migrate(
    direction,
    distance,
    connection_uri,
    database_name,
    path,
    allow_index_dropping,
    use_transaction,
):
    settings_kwargs = {}
    if direction:
        settings_kwargs["direction"] = direction
    if distance:
        settings_kwargs["distance"] = distance
    if connection_uri:
        settings_kwargs["connection_uri"] = connection_uri
    if database_name:
        settings_kwargs["database_name"] = database_name
    if path:
        settings_kwargs["path"] = path
    if allow_index_dropping:
        settings_kwargs["allow_index_dropping"] = allow_index_dropping
    settings_kwargs["use_transaction"] = use_transaction
    settings = MigrationSettings(**settings_kwargs)

    asyncio.run(run_migrate(settings))


@migrations.command()
@click.option("-n", "--name", required=True, type=str, help="Migration name")
@click.option(
    "-p",
    "--path",
    required=True,
    type=str,
    help="Path to the migrations directory",
)
def new_migration(name, path):
    path = Path(path)
    ts_string = datetime.now().strftime("%Y%m%d%H%M%S")
    file_name = f"{ts_string}_{name}.py"

    shutil.copy(template.__file__, path / file_name)


if __name__ == "__main__":
    migrations()



from abc import abstractmethod

from pymongo.asynchronous.collection import AsyncCollection

from beanie.odm.settings.base import ItemSettings


class OtherGettersInterface:
    @classmethod
    @abstractmethod
    def get_settings(cls) -> ItemSettings:
        pass

    @classmethod
    def get_pymongo_collection(cls) -> AsyncCollection:
        return cls.get_settings().pymongo_collection

    @classmethod
    def get_collection_name(cls) -> str:
        return cls.get_settings().name  # type: ignore

    @classmethod
    def get_bson_encoders(cls):
        return cls.get_settings().bson_encoders

    @classmethod
    def get_link_fields(cls):
        return None



import pytest

from beanie.exceptions import (
    DocumentNotFound,
    ReplaceError,
)
from beanie.odm.fields import PydanticObjectId
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import (
    DocumentTestModel,
    DocumentTestModelWithModelConfigExtraAllow,
    DocumentWithKeepNullsFalse,
    DocumentWithList,
    ModelWithOptionalField,
    Sample,
)

# REPLACE

#
# async def test_replace_one(document):
#     new_doc = DocumentTestModel(
#         test_int=0, test_str="REPLACED_VALUE", test_list=[]
#     )
#     await DocumentTestModel.replace_one({"_id": document.id}, new_doc)
#     new_document = await DocumentTestModel.get(document.id)
#     assert new_document.test_str == "REPLACED_VALUE"


async def test_replace_many(documents):
    await documents(10, "foo")
    created_documents = await DocumentTestModel.find_many(
        {"test_str": "foo"}
    ).to_list()
    to_replace = []
    for document in created_documents[:5]:
        document.test_str = "REPLACED_VALUE"
        to_replace.append(document)
    await DocumentTestModel.replace_many(to_replace)

    replaced_documetns = await DocumentTestModel.find_many(
        {"test_str": "REPLACED_VALUE"}
    ).to_list()
    assert len(replaced_documetns) == 5


async def test_replace_many_not_all_the_docs_found(documents):
    await documents(10, "foo")
    created_documents = await DocumentTestModel.find_many(
        {"test_str": "foo"}
    ).to_list()
    to_replace = []
    created_documents[0].id = PydanticObjectId()
    for document in created_documents[:5]:
        document.test_str = "REPLACED_VALUE"
        to_replace.append(document)
    with pytest.raises(ReplaceError):
        await DocumentTestModel.replace_many(to_replace)


async def test_replace(document):
    update_data = {"test_str": "REPLACED_VALUE"}
    if IS_PYDANTIC_V2:
        new_doc = document.model_copy(update=update_data)
    else:
        new_doc = document.copy(update=update_data)
        # pydantic v1 doesn't copy excluded fields
        new_doc.test_list = document.test_list
    # document.test_str = "REPLACED_VALUE"
    await new_doc.replace()
    new_document = await DocumentTestModel.get(document.id)
    assert new_document.test_str == "REPLACED_VALUE"


async def test_replace_not_saved(document_not_inserted):
    with pytest.raises(ValueError):
        await document_not_inserted.replace()


async def test_replace_not_found(document_not_inserted):
    document_not_inserted.id = PydanticObjectId()
    with pytest.raises(DocumentNotFound):
        await document_not_inserted.replace()


# SAVE
async def test_save(document):
    update_data = {"test_str": "REPLACED_VALUE"}
    if IS_PYDANTIC_V2:
        new_doc = document.model_copy(update=update_data)
    else:
        new_doc = document.copy(update=update_data)
    # document.test_str = "REPLACED_VALUE"
    await new_doc.save()
    new_document = await DocumentTestModel.get(document.id)
    assert new_document.test_str == "REPLACED_VALUE"


async def test_save_not_saved(document_not_inserted):
    await document_not_inserted.save()
    assert (
        hasattr(document_not_inserted, "id")
        and document_not_inserted.id is not None
    )
    from_db = await DocumentTestModel.get(document_not_inserted.id)
    assert from_db == document_not_inserted


async def test_save_not_found(document_not_inserted):
    document_not_inserted.id = PydanticObjectId()
    await document_not_inserted.save()
    assert (
        hasattr(document_not_inserted, "id")
        and document_not_inserted.id is not None
    )
    from_db = await DocumentTestModel.get(document_not_inserted.id)
    assert from_db == document_not_inserted


# UPDATE


async def test_update_one(document):
    await DocumentTestModel.find_one(
        {"_id": document.id, "test_list.test_str": "foo"}
    ).update({"$set": {"test_list.$.test_str": "foo_foo"}})
    new_document = await DocumentTestModel.get(document.id)
    assert new_document.test_list[0].test_str == "foo_foo"


async def test_update_one_set_extra_field():
    doc = DocumentTestModelWithModelConfigExtraAllow()
    await doc.insert()

    await doc.update({"$set": {"my_extra_field": 12345}})
    assert doc.my_extra_field == 12345


async def test_update_many(documents):
    await documents(10, "foo")
    await documents(7, "bar")
    await DocumentTestModel.find_many({"test_str": "foo"}).update(
        {"$set": {"test_str": "bar"}}
    )
    bar_documetns = await DocumentTestModel.find_many(
        {"test_str": "bar"}
    ).to_list()
    assert len(bar_documetns) == 17
    foo_documetns = await DocumentTestModel.find_many(
        {"test_str": "foo"}
    ).to_list()
    assert len(foo_documetns) == 0


async def test_update_all(documents):
    await documents(10, "foo")
    await documents(7, "bar")
    await DocumentTestModel.update_all(
        {"$set": {"test_str": "smth_else"}},
    )
    bar_documetns = await DocumentTestModel.find_many(
        {"test_str": "bar"}
    ).to_list()
    assert len(bar_documetns) == 0
    foo_documetns = await DocumentTestModel.find_many(
        {"test_str": "foo"}
    ).to_list()
    assert len(foo_documetns) == 0
    smth_else_documetns = await DocumentTestModel.find_many(
        {"test_str": "smth_else"}
    ).to_list()
    assert len(smth_else_documetns) == 17


async def test_save_keep_nulls_false():
    model = ModelWithOptionalField(i=10, s="TEST_MODEL")
    doc = DocumentWithKeepNullsFalse(m=model, o="TEST_DOCUMENT")

    await doc.insert()

    doc.o = None
    doc.m.s = None
    await doc.save()

    from_db = await DocumentWithKeepNullsFalse.get(doc.id)
    assert from_db.o is None
    assert from_db.m.s is None

    raw_data = (
        await DocumentWithKeepNullsFalse.get_pymongo_collection().find_one(
            {"_id": doc.id}
        )
    )
    assert raw_data == {"_id": doc.id, "m": {"i": 10}}


async def test_save_changes_keep_nulls_false():
    model = ModelWithOptionalField(i=10, s="TEST_MODEL")
    doc = DocumentWithKeepNullsFalse(m=model, o="TEST_DOCUMENT")

    await doc.insert()

    doc.o = None
    doc.m.s = None

    await doc.save_changes()

    from_db = await DocumentWithKeepNullsFalse.get(doc.id)
    assert from_db.o is None
    assert from_db.m.s is None

    raw_data = (
        await DocumentWithKeepNullsFalse.get_pymongo_collection().find_one(
            {"_id": doc.id}
        )
    )
    assert raw_data == {"_id": doc.id, "m": {"i": 10}}


# WITH SESSION


# async def test_update_with_session(document: DocumentTestModel, session):
#     buf_len = len(document.test_list)
#     to_insert = SubDocument(test_str="test")
#     await document.update(
#         update_query={"$push": {"test_list": to_insert.dict()}},
#         session=session,
#     )
#     new_document = await DocumentTestModel.get(document.id, session=session)
#     assert len(new_document.test_list) == buf_len + 1
#
#
# async def test_replace_one_with_session(document, session):
#     new_doc = DocumentTestModel(
#         test_int=0, test_str="REPLACED_VALUE", test_list=[]
#     )
#     await DocumentTestModel.replace_one(
#         {"_id": document.id}, new_doc, session=session
#     )
#     new_document = await DocumentTestModel.get(document.id, session=session)
#     assert new_document.test_str == "REPLACED_VALUE"
#
#
# async def test_replace_with_session(document, session):
#     update_data = {"test_str": "REPLACED_VALUE"}
#     new_doc: DocumentTestModel = document.copy(update=update_data)
#     # document.test_str = "REPLACED_VALUE"
#     await new_doc.replace(session=session)
#     new_document = await DocumentTestModel.get(document.id, session=session)
#     assert new_document.test_str == "REPLACED_VALUE"
#
#
# async def test_update_one_with_session(document, session):
#     await DocumentTestModel.update_one(
#         update_query={"$set": {"test_list.$.test_str": "foo_foo"}},
#         filter_query={"_id": document.id, "test_list.test_str": "foo"},
#         session=session,
#     )
#     new_document = await DocumentTestModel.get(document.id, session=session)
#     assert new_document.test_list[0].test_str == "foo_foo"
#
#
# async def test_update_many_with_session(documents, session):
#     await documents(10, "foo")
#     await documents(7, "bar")
#     await DocumentTestModel.update_many(
#         update_query={"$set": {"test_str": "bar"}},
#         filter_query={"test_str": "foo"},
#         session=session,
#     )
#     bar_documetns = await DocumentTestModel.find_many(
#         {"test_str": "bar"}, session=session
#     ).to_list()
#     assert len(bar_documetns) == 17
#     foo_documetns = await DocumentTestModel.find_many(
#         {"test_str": "foo"}, session=session
#     ).to_list()
#     assert len(foo_documetns) == 0
#
#
# async def test_update_all_with_session(documents, session):
#     await documents(10, "foo")
#     await documents(7, "bar")
#     await DocumentTestModel.update_all(
#         update_query={"$set": {"test_str": "smth_else"}}, session=session
#     )
#     bar_documetns = await DocumentTestModel.find_many(
#         {"test_str": "bar"}, session=session
#     ).to_list()
#     assert len(bar_documetns) == 0
#     foo_documetns = await DocumentTestModel.find_many(
#         {"test_str": "foo"}, session=session
#     ).to_list()
#     assert len(foo_documetns) == 0
#     smth_else_documetns = await DocumentTestModel.find_many(
#         {"test_str": "smth_else"}, session=session
#     ).to_list()
#     assert len(smth_else_documetns) == 17


async def test_update_list():
    test_record = DocumentWithList(list_values=["1", "2", "3"])
    test_record = await test_record.insert()
    if IS_PYDANTIC_V2:
        update_data = test_record.model_dump()
    else:
        update_data = test_record.dict()
    update_data["list_values"] = ["5", "6", "7"]

    updated_test_record = await test_record.update({"$set": update_data})
    assert updated_test_record.list_values == update_data["list_values"]


async def test_update_using_pipeline(preset_documents):
    await Sample.all().update(
        [{"$set": {"integer": 10000}}, {"$set": {"string": "TEST3"}}]
    )
    all_docs = await Sample.find_many({}).to_list()
    for doc in all_docs:
        assert doc.integer == 10000
        assert doc.string == "TEST3"



from typing import Any, Callable, Coroutine

from typing_extensions import Protocol, TypeAlias, assert_type

from beanie import Document
from beanie.odm.actions import EventTypes, wrap_with_actions
from beanie.odm.utils.self_validation import validate_self_before
from beanie.odm.utils.state import (
    previous_saved_state_needed,
    save_state_after,
    saved_state_needed,
)


def sync_func(doc_self: Document, arg1: str, arg2: int, /) -> Document:
    """
    Models `Document` sync method that expects self
    """
    raise NotImplementedError


SyncFunc: TypeAlias = Callable[[Document, str, int], Document]


async def async_func(doc_self: Document, arg1: str, arg2: int, /) -> Document:
    """
    Models `Document` async method that expects self
    """
    raise NotImplementedError


AsyncFunc: TypeAlias = Callable[
    [Document, str, int], Coroutine[Any, Any, Document]
]


def test_wrap_with_actions_preserves_signature() -> None:
    assert_type(async_func, AsyncFunc)
    assert_type(wrap_with_actions(EventTypes.SAVE)(async_func), AsyncFunc)


def test_save_state_after_preserves_signature() -> None:
    assert_type(async_func, AsyncFunc)
    assert_type(save_state_after(async_func), AsyncFunc)


def test_validate_self_before_preserves_signature() -> None:
    assert_type(async_func, AsyncFunc)
    assert_type(validate_self_before(async_func), AsyncFunc)


def test_saved_state_needed_preserves_signature() -> None:
    assert_type(async_func, AsyncFunc)
    assert_type(saved_state_needed(async_func), AsyncFunc)

    assert_type(sync_func, SyncFunc)
    assert_type(saved_state_needed(sync_func), SyncFunc)


def test_previous_saved_state_needed_preserves_signature() -> None:
    assert_type(async_func, AsyncFunc)
    assert_type(previous_saved_state_needed(async_func), AsyncFunc)

    assert_type(sync_func, SyncFunc)
    assert_type(previous_saved_state_needed(sync_func), SyncFunc)


class ExpectsDocumentSelf(Protocol):
    def __call__(self, doc_self: Document, /) -> Any: ...


def test_document_insert_expects_self() -> None:
    test_insert: ExpectsDocumentSelf = Document.insert  # noqa: F841


def test_document_save_expects_self() -> None:
    test_insert: ExpectsDocumentSelf = Document.save  # noqa: F841


def test_document_replace_expects_self() -> None:
    test_insert: ExpectsDocumentSelf = Document.replace  # noqa: F841



from pydantic.main import BaseModel

from beanie import Document, iterative_migration


class OldTag(BaseModel):
    color: str
    name: str


class Tag(BaseModel):
    color: str
    title: str


class OldNote(Document):
    title: str
    tag: OldTag

    class Settings:
        name = "notes"


class Note(Document):
    title: str
    tag: Tag

    class Settings:
        name = "notes"


class Forward:
    @iterative_migration()
    async def change_color(
        self, input_document: OldNote, output_document: Note
    ):
        output_document.tag.title = input_document.tag.name


class Backward:
    @iterative_migration()
    async def change_title(
        self, input_document: Note, output_document: OldNote
    ):
        output_document.tag.name = input_document.tag.title



from tests.odm.models import Sample


async def test_sum(preset_documents, session):
    n = await Sample.find_many(Sample.integer == 1).sum(Sample.increment)

    assert n == 12

    n = await Sample.find_many(Sample.integer == 1).sum(
        Sample.increment, session=session
    )

    assert n == 12


async def test_sum_without_docs(session):
    n = await Sample.find_many(Sample.integer == 1).sum(Sample.increment)

    assert n is None

    n = await Sample.find_many(Sample.integer == 1).sum(
        Sample.increment, session=session
    )

    assert n is None


async def test_avg(preset_documents, session):
    n = await Sample.find_many(Sample.integer == 1).avg(Sample.increment)

    assert n == 4
    n = await Sample.find_many(Sample.integer == 1).avg(
        Sample.increment, session=session
    )

    assert n == 4


async def test_avg_without_docs(session):
    n = await Sample.find_many(Sample.integer == 1).avg(Sample.increment)

    assert n is None
    n = await Sample.find_many(Sample.integer == 1).avg(
        Sample.increment, session=session
    )

    assert n is None


async def test_max(preset_documents, session):
    n = await Sample.find_many(Sample.integer == 1).max(Sample.increment)

    assert n == 5

    n = await Sample.find_many(Sample.integer == 1).max(
        Sample.increment, session=session
    )

    assert n == 5


async def test_max_without_docs(session):
    n = await Sample.find_many(Sample.integer == 1).max(Sample.increment)

    assert n is None

    n = await Sample.find_many(Sample.integer == 1).max(
        Sample.increment, session=session
    )

    assert n is None


async def test_min(preset_documents, session):
    n = await Sample.find_many(Sample.integer == 1).min(Sample.increment)

    assert n == 3

    n = await Sample.find_many(Sample.integer == 1).min(
        Sample.increment, session=session
    )

    assert n == 3


async def test_min_without_docs(session):
    n = await Sample.find_many(Sample.integer == 1).min(Sample.increment)

    assert n is None

    n = await Sample.find_many(Sample.integer == 1).min(
        Sample.increment, session=session
    )

    assert n is None


async def test_all_sum(preset_documents, session):
    n = await Sample.sum(Sample.increment)

    assert n == 45

    n = await Sample.sum(Sample.increment, session=session)

    assert n == 45


async def test_all_sum_without_docs(session):
    n = await Sample.sum(Sample.increment)

    assert n is None

    n = await Sample.sum(Sample.increment, session=session)

    assert n is None


async def test_all_avg(preset_documents, session):
    n = await Sample.avg(Sample.increment)

    assert n == 4.5
    n = await Sample.avg(Sample.increment, session=session)

    assert n == 4.5


async def test_all_avg_without_docs(session):
    n = await Sample.avg(Sample.increment)

    assert n is None
    n = await Sample.avg(Sample.increment, session=session)

    assert n is None


async def test_all_max(preset_documents, session):
    n = await Sample.max(Sample.increment)

    assert n == 9

    n = await Sample.max(Sample.increment, session=session)

    assert n == 9


async def test_all_max_without_docs(session):
    n = await Sample.max(Sample.increment)

    assert n is None

    n = await Sample.max(Sample.increment, session=session)

    assert n is None


async def test_all_min(preset_documents, session):
    n = await Sample.min(Sample.increment)

    assert n == 0

    n = await Sample.min(Sample.increment, session=session)

    assert n == 0


async def test_all_min_without_docs(session):
    n = await Sample.min(Sample.increment)

    assert n is None

    n = await Sample.min(Sample.increment, session=session)

    assert n is None



from typing import Any, ClassVar, Dict, Optional, Type, TypeVar

from pymongo.asynchronous.client_session import AsyncClientSession

from beanie.exceptions import UnionDocNotInited
from beanie.odm.bulk import BulkWriter
from beanie.odm.interfaces.aggregate import AggregateInterface
from beanie.odm.interfaces.detector import DetectionInterface, ModelType
from beanie.odm.interfaces.find import FindInterface
from beanie.odm.interfaces.getters import OtherGettersInterface
from beanie.odm.settings.union_doc import UnionDocSettings

UnionDocType = TypeVar("UnionDocType", bound="UnionDoc")


class UnionDoc(
    FindInterface,
    AggregateInterface,
    OtherGettersInterface,
    DetectionInterface,
):
    _document_models: ClassVar[Optional[Dict[str, Type]]] = None
    _is_inited: ClassVar[bool] = False
    _settings: ClassVar[UnionDocSettings]

    @classmethod
    def get_settings(cls) -> UnionDocSettings:
        return cls._settings

    @classmethod
    def register_doc(cls, name: str, doc_model: Type):
        if cls._document_models is None:
            cls._document_models = {}

        if cls._is_inited is False:
            raise UnionDocNotInited

        cls._document_models[name] = doc_model
        return cls.get_settings().name

    @classmethod
    def get_model_type(cls) -> ModelType:
        return ModelType.UnionDoc

    @classmethod
    def bulk_writer(
        cls,
        session: Optional[AsyncClientSession] = None,
        ordered: bool = True,
        bypass_document_validation: bool = False,
        comment: Optional[Any] = None,
    ) -> BulkWriter:
        """
        Returns a BulkWriter instance for handling bulk write operations.

        :param session: Optional[AsyncClientSession] - pymongo session.
            The session instance used for transactional operations.
        :param ordered: bool
            If ``True`` (the default), requests will be performed on the server serially, in the order provided. If an error
            occurs, all remaining operations are aborted. If ``False``, requests will be performed on the server in
            arbitrary order, possibly in parallel, and all operations will be attempted.
        :param bypass_document_validation: bool, optional
            If ``True``, allows the write to opt-out of document-level validation. Default is ``False``.
        :param comment: str, optional
            A user-provided comment to attach to the BulkWriter.

        :returns: BulkWriter
            An instance of BulkWriter configured with the provided settings.

        Example Usage:
        --------------
        This method is typically used within an asynchronous context manager.

        .. code-block:: python

            async with Document.bulk_writer(ordered=True) as bulk:
                await Document.insert_one(Document(field="value"), bulk_writer=bulk)
        """
        return BulkWriter(
            session, ordered, cls, bypass_document_validation, comment
        )



from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import DocumentWithRootModelAsAField

if IS_PYDANTIC_V2:

    class TestRootModels:
        async def test_insert(self):
            doc = DocumentWithRootModelAsAField(pets=["dog", "cat", "fish"])
            await doc.insert()

            new_doc = await DocumentWithRootModelAsAField.get(doc.id)
            assert new_doc.pets.root == ["dog", "cat", "fish"]

            collection = DocumentWithRootModelAsAField.get_pymongo_collection()
            raw_doc = await collection.find_one({"_id": doc.id})
            assert raw_doc["pets"] == ["dog", "cat", "fish"]



from beanie.odm.operators.find.element import Exists, Type
from tests.odm.models import Sample


async def test_exists():
    q = Exists(Sample.integer, True)
    assert q == {"integer": {"$exists": True}}

    q = Exists(Sample.integer, False)
    assert q == {"integer": {"$exists": False}}

    q = Exists(Sample.integer)
    assert q == {"integer": {"$exists": True}}


async def test_type():
    q = Type(Sample.integer, "smth")
    assert q == {"integer": {"$type": "smth"}}



from typing import Any, Dict, List, Type, Union

from pydantic import Field

from beanie.odm.settings.base import ItemSettings


class ViewSettings(ItemSettings):
    source: Union[str, Type]
    pipeline: List[Dict[str, Any]]

    max_nesting_depths_per_field: dict = Field(default_factory=dict)
    max_nesting_depth: int = 3



import pytest
from pydantic.main import BaseModel

from beanie import init_beanie
from beanie.executors.migrate import MigrationSettings, run_migrate
from beanie.migrations.models import RunningDirections
from beanie.odm.documents import Document
from beanie.odm.models import InspectionStatuses


class OldTag(BaseModel):
    color: str
    name: str


class Tag(BaseModel):
    color: str
    title: str


class OldNote(Document):
    title: str
    tag: OldTag

    class Settings:
        name = "notes"


class Note(Document):
    title: str
    tag: Tag

    class Settings:
        name = "notes"


@pytest.fixture()
async def notes(db):
    await init_beanie(database=db, document_models=[OldNote])
    await OldNote.delete_all()
    for i in range(10):
        note = OldNote(title=str(i), tag=OldTag(name="test", color="red"))
        await note.insert()
    yield
    await OldNote.delete_all()


async def test_migration_change_subfield_value(settings, notes, db):
    migration_settings = MigrationSettings(
        connection_uri=settings.mongodb_dsn,
        database_name=settings.mongodb_db_name,
        path="tests/migrations/migrations_for_test/change_subfield",
    )
    await run_migrate(migration_settings)

    await init_beanie(database=db, document_models=[Note])
    inspection = await Note.inspect_collection()
    assert inspection.status == InspectionStatuses.OK
    note = await Note.find_one({})
    assert note.tag.title == "test"

    migration_settings.direction = RunningDirections.BACKWARD
    await run_migrate(migration_settings)
    inspection = await OldNote.inspect_collection()
    assert inspection.status == InspectionStatuses.OK
    note = await OldNote.find_one({})
    assert note.tag.name == "test"



import sys

from pymongo.asynchronous.database import AsyncDatabase
from typing_extensions import Sequence, get_args, get_origin

from beanie.odm.utils.pydantic import (
    IS_PYDANTIC_V2,
    get_extra_field_info,
    get_model_fields,
    parse_model,
)
from beanie.odm.utils.typing import get_index_attributes

if sys.version_info >= (3, 10):
    from types import UnionType as TypesUnionType
else:
    TypesUnionType = ()

import importlib
import inspect
from typing import (  # type: ignore
    List,
    Optional,
    Type,
    Union,
    _GenericAlias,
)

from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pymongo import AsyncMongoClient, IndexModel

from beanie.exceptions import Deprecation, MongoDBVersionError
from beanie.odm.actions import ActionRegistry
from beanie.odm.cache import LRUCache
from beanie.odm.documents import DocType, Document
from beanie.odm.fields import (
    BackLink,
    ExpressionField,
    Link,
    LinkInfo,
    LinkTypes,
)
from beanie.odm.interfaces.detector import ModelType
from beanie.odm.registry import DocsRegistry
from beanie.odm.settings.document import DocumentSettings, IndexModelField
from beanie.odm.settings.union_doc import UnionDocSettings
from beanie.odm.settings.view import ViewSettings
from beanie.odm.union_doc import UnionDoc, UnionDocType
from beanie.odm.views import View


class Output(BaseModel):
    class_name: str
    collection_name: str


class Initializer:
    def __init__(
        self,
        database: AsyncDatabase = None,
        connection_string: Optional[str] = None,
        document_models: Optional[
            Sequence[
                Union[Type["DocType"], Type["UnionDocType"], Type["View"], str]
            ]
        ] = None,
        allow_index_dropping: bool = False,
        recreate_views: bool = False,
        skip_indexes: bool = False,
    ):
        """
        Beanie initializer

        :param database: AsyncDatabase - pymongo database instance
        :param connection_string: str - MongoDB connection string
        :param document_models: List[Union[Type[DocType], Type[UnionDocType], str]] - model classes
        or strings with dot separated paths
        :param allow_index_dropping: bool - if index dropping is allowed.
        Default False
        :param recreate_views: bool - if views should be recreated. Default False
        it will patch the pymongo client to use process's event loop.
        :param skip_indexes: bool - if you want to skip working with indexes. Default False
        :return: None
        """

        self.inited_classes: List[Type] = []
        self.allow_index_dropping = allow_index_dropping
        self.skip_indexes = skip_indexes
        self.recreate_views = recreate_views

        self.models_with_updated_forward_refs: List[Type[BaseModel]] = []

        if (connection_string is None and database is None) or (
            connection_string is not None and database is not None
        ):
            raise ValueError(
                "connection_string parameter or database parameter must be set"
            )

        if document_models is None:
            raise ValueError("document_models parameter must be set")
        if connection_string is not None:
            database = AsyncMongoClient(
                connection_string
            ).get_default_database()

        self.database: AsyncDatabase = database

        sort_order = {
            ModelType.UnionDoc: 0,
            ModelType.Document: 1,
            ModelType.View: 2,
        }

        self.document_models: List[
            Union[Type[DocType], Type[UnionDocType], Type[View]]
        ] = [
            self.get_model(model) if isinstance(model, str) else model
            for model in document_models
        ]

        self.fill_docs_registry()

        self.document_models.sort(
            key=lambda val: sort_order[val.get_model_type()]
        )

    def __await__(self):
        for model in self.document_models:
            yield from self.init_class(model).__await__()

    # General
    def fill_docs_registry(self):
        for model in self.document_models:
            module = inspect.getmodule(model)
            members = inspect.getmembers(module)
            for name, obj in members:
                if inspect.isclass(obj) and issubclass(obj, BaseModel):
                    DocsRegistry.register(name, obj)

    @staticmethod
    def get_model(dot_path: str) -> Type["DocType"]:
        """
        Get the model by the path in format bar.foo.Model

        :param dot_path: str - dot seprated path to the model
        :return: Type[DocType] - class of the model
        """
        module_name, class_name = None, None
        try:
            module_name, class_name = dot_path.rsplit(".", 1)
            return getattr(importlib.import_module(module_name), class_name)

        except ValueError:
            raise ValueError(
                f"'{dot_path}' doesn't have '.' path, eg. path.to.your.model.class"
            )

        except AttributeError:
            raise AttributeError(
                f"module '{module_name}' has no class called '{class_name}'"
            )

    def init_settings(
        self, cls: Union[Type[Document], Type[View], Type[UnionDoc]]
    ):
        """
        Init Settings

        :param cls: Union[Type[Document], Type[View], Type[UnionDoc]] - Class
        to init settings
        :return: None
        """
        settings_class = getattr(cls, "Settings", None)
        settings_vars = {}
        if settings_class is not None:
            # get all attributes of the Settings subclass (including inherited ones)
            # without magic dunder methods
            settings_vars = {
                attr: getattr(settings_class, attr)
                for attr in dir(settings_class)
                if not attr.startswith("__")
            }
        if issubclass(cls, Document):
            cls._document_settings = parse_model(
                DocumentSettings, settings_vars
            )
        if issubclass(cls, View):
            cls._settings = parse_model(ViewSettings, settings_vars)
        if issubclass(cls, UnionDoc):
            cls._settings = parse_model(UnionDocSettings, settings_vars)

    if not IS_PYDANTIC_V2:

        def update_forward_refs(self, cls: Type[BaseModel]):
            """
            Update forward refs

            :param cls: Type[BaseModel] - class to update forward refs
            :return: None
            """
            if cls not in self.models_with_updated_forward_refs:
                cls.update_forward_refs()
                self.models_with_updated_forward_refs.append(cls)

    # General. Relations

    def detect_link(
        self, field: FieldInfo, field_name: str
    ) -> Optional[LinkInfo]:
        """
        It detects link and returns LinkInfo if any found.

        :param field: ModelField
        :return: Optional[LinkInfo]
        """

        origin = get_origin(field.annotation)
        args = get_args(field.annotation)
        classes = [
            Link,
            BackLink,
        ]

        for cls in classes:
            # Check if annotation is one of the custom classes
            if (
                isinstance(field.annotation, _GenericAlias)
                and field.annotation.__origin__ is cls
            ):
                if cls is Link:
                    return LinkInfo(
                        field_name=field_name,
                        lookup_field_name=field_name,
                        document_class=DocsRegistry.evaluate_fr(args[0]),  # type: ignore
                        link_type=LinkTypes.DIRECT,
                    )
                if cls is BackLink:
                    return LinkInfo(
                        field_name=field_name,
                        lookup_field_name=get_extra_field_info(
                            field, "original_field"
                        ),  # type: ignore
                        document_class=DocsRegistry.evaluate_fr(args[0]),  # type: ignore
                        link_type=LinkTypes.BACK_DIRECT,
                    )

            # Check if annotation is List[custom class]
            elif (
                (origin is List or origin is list)
                and len(args) == 1
                and isinstance(args[0], _GenericAlias)
                and args[0].__origin__ is cls
            ):
                if cls is Link:
                    return LinkInfo(
                        field_name=field_name,
                        lookup_field_name=field_name,
                        document_class=DocsRegistry.evaluate_fr(
                            get_args(args[0])[0]
                        ),  # type: ignore
                        link_type=LinkTypes.LIST,
                    )
                if cls is BackLink:
                    return LinkInfo(
                        field_name=field_name,
                        lookup_field_name=get_extra_field_info(  # type: ignore
                            field, "original_field"
                        ),
                        document_class=DocsRegistry.evaluate_fr(
                            get_args(args[0])[0]
                        ),  # type: ignore
                        link_type=LinkTypes.BACK_LIST,
                    )

            # Check if annotation is Optional[custom class] or Optional[List[custom class]]
            elif (
                (origin is Union or origin is TypesUnionType)
                and len(args) == 2
                and type(None) in args
            ):
                if args[1] is type(None):
                    optional = args[0]
                else:
                    optional = args[1]
                optional_origin = get_origin(optional)
                optional_args = get_args(optional)

                if (
                    isinstance(optional, _GenericAlias)
                    and optional.__origin__ is cls
                ):
                    if cls is Link:
                        return LinkInfo(
                            field_name=field_name,
                            lookup_field_name=field_name,
                            document_class=DocsRegistry.evaluate_fr(
                                optional_args[0]
                            ),  # type: ignore
                            link_type=LinkTypes.OPTIONAL_DIRECT,
                        )
                    if cls is BackLink:
                        return LinkInfo(
                            field_name=field_name,
                            lookup_field_name=get_extra_field_info(
                                field, "original_field"
                            ),
                            document_class=DocsRegistry.evaluate_fr(
                                optional_args[0]
                            ),  # type: ignore
                            link_type=LinkTypes.OPTIONAL_BACK_DIRECT,
                        )

                elif (
                    (optional_origin is List or optional_origin is list)
                    and len(optional_args) == 1
                    and isinstance(optional_args[0], _GenericAlias)
                    and optional_args[0].__origin__ is cls
                ):
                    if cls is Link:
                        return LinkInfo(
                            field_name=field_name,
                            lookup_field_name=field_name,
                            document_class=DocsRegistry.evaluate_fr(
                                get_args(optional_args[0])[0]
                            ),  # type: ignore
                            link_type=LinkTypes.OPTIONAL_LIST,
                        )
                    if cls is BackLink:
                        return LinkInfo(
                            field_name=field_name,
                            lookup_field_name=get_extra_field_info(
                                field, "original_field"
                            ),
                            document_class=DocsRegistry.evaluate_fr(
                                get_args(optional_args[0])[0]
                            ),  # type: ignore
                            link_type=LinkTypes.OPTIONAL_BACK_LIST,
                        )
        return None

    def check_nested_links(self, link_info: LinkInfo, current_depth: int):
        if current_depth == 1:
            return
        for k, v in get_model_fields(link_info.document_class).items():
            nested_link_info = self.detect_link(v, k)
            if nested_link_info is None:
                continue

            if link_info.nested_links is None:
                link_info.nested_links = {}
            link_info.nested_links[k] = nested_link_info
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            self.check_nested_links(nested_link_info, current_depth=new_depth)

    # Document

    @staticmethod
    def set_default_class_vars(cls: Type[Document]):
        """
        Set default class variables.

        :param cls: Union[Type[Document], Type[View], Type[UnionDoc]] - Class
        to init settings
        :return:
        """
        cls._children = dict()
        cls._parent = None
        cls._inheritance_inited = False
        cls._class_id = None
        cls._link_fields = None

    @staticmethod
    def init_cache(cls) -> None:
        """
        Init model's cache
        :return: None
        """
        if cls.get_settings().use_cache:
            cls._cache = LRUCache(
                capacity=cls.get_settings().cache_capacity,
                expiration_time=cls.get_settings().cache_expiration_time,
            )

    def init_document_fields(self, cls) -> None:
        """
        Init class fields
        :return: None
        """

        if not IS_PYDANTIC_V2:
            self.update_forward_refs(cls)

        if cls._link_fields is None:
            cls._link_fields = {}
        for k, v in get_model_fields(cls).items():
            path = v.alias or k
            setattr(cls, k, ExpressionField(path))

            link_info = self.detect_link(v, k)
            depth_level = cls.get_settings().max_nesting_depths_per_field.get(
                k, None
            )
            if depth_level is None:
                depth_level = cls.get_settings().max_nesting_depth
            if link_info is not None:
                if depth_level > 0 or depth_level is None:
                    cls._link_fields[k] = link_info
                    self.check_nested_links(
                        link_info, current_depth=depth_level
                    )
                elif depth_level <= 0:
                    link_info.is_fetchable = False
                    cls._link_fields[k] = link_info

        cls._check_hidden_fields()

    @staticmethod
    def init_actions(cls):
        """
        Init event-based actions
        """
        ActionRegistry.clean_actions(cls)
        for attr in dir(cls):
            f = getattr(cls, attr)
            if inspect.isfunction(f):
                if hasattr(f, "has_action"):
                    ActionRegistry.add_action(
                        document_class=cls,
                        event_types=f.event_types,  # type: ignore
                        action_direction=f.action_direction,  # type: ignore
                        funct=f,
                    )

    async def init_document_collection(self, cls):
        """
        Init collection for the Document-based class
        :param cls:
        :return:
        """
        cls.set_database(self.database)

        document_settings = cls.get_settings()

        # register in the Union Doc

        if document_settings.union_doc is not None:
            name = cls.get_settings().name or cls.__name__
            document_settings.name = document_settings.union_doc.register_doc(
                name, cls
            )
            document_settings.union_doc_alias = name

        # set a name

        if not document_settings.name:
            document_settings.name = cls.__name__

        # check mongodb version fits
        if (
            document_settings.timeseries is not None
            and cls._database_major_version < 5
        ):
            raise MongoDBVersionError(
                "Timeseries are supported by MongoDB version 5 and higher"
            )

        # create pymongo collection
        if (
            document_settings.timeseries is not None
            and document_settings.name
            not in await self.database.list_collection_names(
                authorizedCollections=True, nameOnly=True
            )
        ):
            collection = await self.database.create_collection(
                **document_settings.timeseries.build_query(
                    document_settings.name
                )
            )
        else:
            collection = self.database[document_settings.name]

        cls.set_collection(collection)

    async def init_indexes(self, cls, allow_index_dropping: bool = False):
        """
        Async indexes initializer
        """
        collection = cls.get_pymongo_collection()
        document_settings = cls.get_settings()

        index_information = await collection.index_information()

        old_indexes = IndexModelField.from_pymongo_index_information(
            index_information
        )
        new_indexes = []

        # Indexed field wrapped with Indexed()
        indexed_fields = (
            (k, fvalue, get_index_attributes(fvalue))
            for k, fvalue in get_model_fields(cls).items()
        )
        found_indexes = [
            IndexModelField(
                IndexModel(
                    [
                        (
                            fvalue.alias or k,
                            indexed_attrs[0],
                        )
                    ],
                    **indexed_attrs[1],
                )
            )
            for k, fvalue, indexed_attrs in indexed_fields
            if indexed_attrs is not None
        ]

        if document_settings.merge_indexes:
            result: List[IndexModelField] = []
            for subclass in reversed(cls.mro()):
                if issubclass(subclass, Document) and not subclass == Document:
                    if (
                        subclass not in self.inited_classes
                        and not subclass == cls
                    ):
                        await self.init_class(subclass)
                    if subclass.get_settings().indexes:
                        result = IndexModelField.merge_indexes(
                            result, subclass.get_settings().indexes
                        )
            found_indexes = IndexModelField.merge_indexes(
                found_indexes, result
            )

        else:
            if document_settings.indexes:
                found_indexes = IndexModelField.merge_indexes(
                    found_indexes, document_settings.indexes
                )

        new_indexes += found_indexes

        # delete indexes
        # Only drop indexes if the user specifically allows for it
        if allow_index_dropping:
            for index in IndexModelField.list_difference(
                old_indexes, new_indexes
            ):
                await collection.drop_index(index.name)

        # create indices
        if found_indexes:
            new_indexes += await collection.create_indexes(
                IndexModelField.list_to_index_model(new_indexes)
            )

    async def init_document(self, cls: Type[Document]) -> Optional[Output]:
        """
        Init Document-based class

        :param cls:
        :return:
        """
        if cls is Document:
            return None

        # get db version
        build_info = await self.database.command({"buildInfo": 1})
        mongo_version = build_info["version"]
        cls._database_major_version = int(mongo_version.split(".")[0])
        if cls not in self.inited_classes:
            self.set_default_class_vars(cls)
            self.init_settings(cls)

            bases = [b for b in cls.__bases__ if issubclass(b, Document)]
            if len(bases) > 1:
                return None
            parent = bases[0]
            output = await self.init_document(parent)
            if cls.get_settings().is_root and (
                parent is Document or not parent.get_settings().is_root
            ):
                if cls.get_collection_name() is None:
                    cls.set_collection_name(cls.__name__)
                output = Output(
                    class_name=cls.__name__,
                    collection_name=cls.get_collection_name(),
                )
                cls._class_id = cls.__name__
                cls._inheritance_inited = True
            elif output is not None:
                output.class_name = f"{output.class_name}.{cls.__name__}"
                cls._class_id = output.class_name
                cls.set_collection_name(output.collection_name)
                parent.add_child(cls._class_id, cls)
                cls._parent = parent
                cls._inheritance_inited = True

            await self.init_document_collection(cls)
            if not self.skip_indexes:
                await self.init_indexes(cls, self.allow_index_dropping)
            self.init_document_fields(cls)
            self.init_cache(cls)
            self.init_actions(cls)

            self.inited_classes.append(cls)

            return output

        else:
            if cls._inheritance_inited is True:
                return Output(
                    class_name=cls._class_id,
                    collection_name=cls.get_collection_name(),
                )
            else:
                return None

    # Views

    def init_view_fields(self, cls) -> None:
        """
        Init class fields
        :return: None
        """

        if cls._link_fields is None:
            cls._link_fields = {}
        for k, v in get_model_fields(cls).items():
            path = v.alias or k
            setattr(cls, k, ExpressionField(path))
            link_info = self.detect_link(v, k)
            depth_level = cls.get_settings().max_nesting_depths_per_field.get(
                k, None
            )
            if depth_level is None:
                depth_level = cls.get_settings().max_nesting_depth
            if link_info is not None:
                if depth_level > 0:
                    cls._link_fields[k] = link_info
                    self.check_nested_links(
                        link_info, current_depth=depth_level
                    )
                elif depth_level <= 0:
                    link_info.is_fetchable = False
                    cls._link_fields[k] = link_info

    def init_view_collection(self, cls):
        """
        Init collection for View

        :param cls:
        :return:
        """
        view_settings = cls.get_settings()

        if view_settings.name is None:
            view_settings.name = cls.__name__

        if inspect.isclass(view_settings.source):
            view_settings.source = view_settings.source.get_collection_name()

        view_settings.pymongo_db = self.database
        view_settings.pymongo_collection = self.database[view_settings.name]

    async def init_view(self, cls: Type[View]):
        """
        Init View-based class

        :param cls:
        :return:
        """
        self.init_settings(cls)
        self.init_view_collection(cls)
        self.init_view_fields(cls)
        self.init_cache(cls)

        collection_names = await self.database.list_collection_names(
            authorizedCollections=True, nameOnly=True
        )
        if self.recreate_views or cls._settings.name not in collection_names:
            if cls._settings.name in collection_names:
                await cls.get_pymongo_collection().drop()

            await self.database.command(
                {
                    "create": cls.get_settings().name,
                    "viewOn": cls.get_settings().source,
                    "pipeline": cls.get_settings().pipeline,
                }
            )

    # Union Doc

    async def init_union_doc(self, cls: Type[UnionDoc]):
        """
        Init Union Doc based class

        :param cls:
        :return:
        """
        self.init_settings(cls)
        if cls._settings.name is None:
            cls._settings.name = cls.__name__

        cls._settings.pymongo_db = self.database
        cls._settings.pymongo_collection = self.database[cls._settings.name]
        cls._is_inited = True

    # Deprecations

    @staticmethod
    def check_deprecations(
        cls: Union[Type[Document], Type[View], Type[UnionDoc]],
    ):
        if hasattr(cls, "Collection"):
            raise Deprecation(
                "Collection inner class is not supported more. "
                "Please use Settings instead. "
                "https://beanie-odm.dev/tutorial/defining-a-document/#settings"
            )

    # Final

    async def init_class(
        self, cls: Union[Type[Document], Type[View], Type[UnionDoc]]
    ):
        """
        Init Document, View or UnionDoc based class.

        :param cls:
        :return:
        """
        self.check_deprecations(cls)

        if issubclass(cls, Document):
            await self.init_document(cls)

        if issubclass(cls, View):
            await self.init_view(cls)

        if issubclass(cls, UnionDoc):
            await self.init_union_doc(cls)

        if hasattr(cls, "custom_init"):
            await cls.custom_init()  # type: ignore


async def init_beanie(
    database: Optional[AsyncDatabase] = None,
    connection_string: Optional[str] = None,
    document_models: Optional[
        Sequence[Union[Type[Document], Type[UnionDoc], Type["View"], str]]
    ] = None,
    allow_index_dropping: bool = False,
    recreate_views: bool = False,
    skip_indexes: bool = False,
):
    """
    Beanie initialization

    :param database: Optional[AsyncDatabase] - pymongo database instance. Defaults to None.
    :param connection_string: Optional[str] - MongoDB connection string.  Defaults to None.
    :param document_models: List[Union[Type[DocType], Type[UnionDocType], str]] - model classes
        or strings with dot separated paths. Defaults to None.
    :param allow_index_dropping: bool - if index dropping is allowed.
        Defaults to False.
    :param recreate_views: bool - if views should be recreated. Defaults to False.
    :param skip_indexes: bool - if you want to skip working with the indexes.
        Defaults to False.
    :return: None
    """

    await Initializer(
        database=database,
        connection_string=connection_string,
        document_models=document_models,
        allow_index_dropping=allow_index_dropping,
        recreate_views=recreate_views,
        skip_indexes=skip_indexes,
    )



import subprocess
import sys
from pathlib import Path

import requests  # type: ignore

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
import tomli_w
from gh import GitHubHandler


class SemVer:
    def __init__(self, version: str):
        self.version = version
        self.major, self.minor, self.patch = map(int, self.version.split("."))

    def increment_minor(self):
        return SemVer(f"{self.major}.{self.minor + 1}.0")

    def __str__(self):
        return self.version

    def __eq__(self, other):
        return self.version == other.version

    def __gt__(self, other):
        return (
            (self.major > other.major)
            or (self.major == other.major and self.minor > other.minor)
            or (
                self.major == other.major
                and self.minor == other.minor
                and self.patch > other.patch
            )
        )


class VersionHandler:
    PACKAGE_NAME = "beanie"
    ROOT_PATH = Path(__file__).parent.parent.parent.parent

    def __init__(self):
        self.pyproject = self.ROOT_PATH / "pyproject.toml"
        self.init_py = self.ROOT_PATH / "beanie" / "__init__.py"
        self.changelog = self.ROOT_PATH / "docs" / "changelog.md"

        self.current_version = self.parse_version_from_pyproject(
            self.pyproject
        )
        self.pypi_version = self.get_version_from_pypi()

        if self.current_version < self.pypi_version:
            raise ValueError("Current version is less than pypi version")

        if self.current_version == self.pypi_version:
            self.current_version = self.current_version.increment_minor()
            self.update_files()
        else:
            self.flit_publish()

    @staticmethod
    def parse_version_from_pyproject(pyproject: Path) -> SemVer:
        with pyproject.open("rb") as f:
            toml_data = tomllib.load(f)
        return SemVer(toml_data["project"]["version"])

    def get_version_from_pypi(self) -> SemVer:
        response = requests.get(
            f"https://pypi.org/pypi/{self.PACKAGE_NAME}/json"
        )
        if response.status_code == 200:
            return SemVer(response.json()["info"]["version"])
        raise ValueError("Can't get version from pypi")

    def update_files(self):
        self.update_pyproject_version()
        self.update_file_versions([self.init_py])
        self.update_changelog()

    def update_pyproject_version(self):
        with self.pyproject.open("rb") as f:
            pyproject = tomllib.load(f)
        pyproject["project"]["version"] = str(self.current_version)
        with self.pyproject.open("wb") as f:
            tomli_w.dump(pyproject, f)

    def update_file_versions(self, files_to_update):
        for file_path in files_to_update:
            content = file_path.read_text()
            content = content.replace(
                str(self.pypi_version), str(self.current_version)
            )
            file_path.write_text(content)

    def update_changelog(self):
        handler = GitHubHandler(
            "BeanieODM",
            "beanie",
            str(self.pypi_version),
            str(self.current_version),
        )
        changelog_content = handler.build_markdown_for_many_prs()

        changelog_lines = self.changelog.read_text().splitlines()
        new_changelog_lines = []
        inserted = False

        for line in changelog_lines:
            new_changelog_lines.append(line)
            if line.strip() == "# Changelog" and not inserted:
                new_changelog_lines.append(changelog_content)
                inserted = True

        self.changelog.write_text("\n".join(new_changelog_lines))
        handler.commit_changes()

    def flit_publish(self):
        subprocess.run(["flit", "publish"], check=True)


if __name__ == "__main__":
    VersionHandler()



from uuid import UUID

import pytest
from pydantic import BaseModel

from beanie import PydanticObjectId
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import DocumentWithCustomIdInt, DocumentWithCustomIdUUID


class A(BaseModel):
    id: PydanticObjectId


async def test_uuid_id():
    doc = DocumentWithCustomIdUUID(name="TEST")
    await doc.insert()
    new_doc = await DocumentWithCustomIdUUID.get(doc.id)
    assert isinstance(new_doc.id, UUID)


async def test_integer_id():
    doc = DocumentWithCustomIdInt(name="TEST", id=1)
    await doc.insert()
    new_doc = await DocumentWithCustomIdInt.get(doc.id)
    assert isinstance(new_doc.id, int)


@pytest.mark.skipif(
    not IS_PYDANTIC_V2,
    reason="supports only pydantic v2",
)
async def test_pydantic_object_id_validation_json():
    deserialized = A.model_validate_json('{"id": "5eb7cf5a86d9755df3a6c593"}')
    assert isinstance(deserialized.id, PydanticObjectId)
    assert str(deserialized.id) == "5eb7cf5a86d9755df3a6c593"
    assert deserialized.id == PydanticObjectId("5eb7cf5a86d9755df3a6c593")


@pytest.mark.skipif(
    not IS_PYDANTIC_V2,
    reason="supports only pydantic v2",
)
@pytest.mark.parametrize(
    "data",
    [
        "5eb7cf5a86d9755df3a6c593",
        PydanticObjectId("5eb7cf5a86d9755df3a6c593"),
    ],
)
async def test_pydantic_object_id_serialization(data):
    deserialized = A(**{"id": data})
    assert isinstance(deserialized.id, PydanticObjectId)
    assert str(deserialized.id) == "5eb7cf5a86d9755df3a6c593"
    assert deserialized.id == PydanticObjectId("5eb7cf5a86d9755df3a6c593")



import datetime
from enum import Enum

import pytest
from pydantic import BaseModel

from beanie.odm.enums import SortDirection
from tests.odm.models import (
    Color,
    DocumentWithBsonEncodersFiledsTypes,
    Door,
    House,
    Lock,
    Sample,
    Window,
)


async def test_find_query():
    q = Sample.find_many(Sample.integer == 1).get_filter_query()
    assert q == {"integer": 1}

    q = Sample.find_many(
        Sample.integer == 1, Sample.nested.integer >= 2
    ).get_filter_query()
    assert q == {"$and": [{"integer": 1}, {"nested.integer": {"$gte": 2}}]}

    q = (
        Sample.find_many(Sample.integer == 1)
        .find_many(Sample.nested.integer >= 2)
        .get_filter_query()
    )
    assert q == {"$and": [{"integer": 1}, {"nested.integer": {"$gte": 2}}]}

    q = Sample.find().get_filter_query()
    assert q == {}


async def test_find_many(preset_documents):
    result = (
        await Sample.find_many(Sample.integer > 1)
        .find_many(Sample.nested.optional == None)
        .to_list()
    )  # noqa
    assert len(result) == 2
    for a in result:
        assert a.integer > 1
        assert a.nested.optional is None

    len_result = 0
    async for a in Sample.find_many(Sample.integer > 1).find_many(
        Sample.nested.optional == None
    ):  # noqa
        assert a in result
        len_result += 1

    assert len_result == len(result)


async def test_find_many_skip(preset_documents):
    q = Sample.find_many(Sample.integer > 1, skip=2)
    assert q.skip_number == 2

    q = Sample.find_many(Sample.integer > 1).skip(2)
    assert q.skip_number == 2

    result = (
        await Sample.find_many(Sample.increment > 2)
        .find_many(Sample.nested.optional == None)
        .skip(1)
        .to_list()
    )
    assert len(result) == 3
    for sample in result:
        assert sample.increment > 2
        assert sample.nested.optional is None

    len_result = 0
    async for sample in (
        Sample.find_many(Sample.increment > 2)
        .find_many(Sample.nested.optional == None)
        .skip(1)
    ):  # noqa
        assert sample in result
        len_result += 1

    assert len_result == len(result)


async def test_find_many_limit(preset_documents):
    q = Sample.find_many(Sample.integer > 1, limit=2)
    assert q.limit_number == 2

    q = Sample.find_many(Sample.integer > 1).limit(2)
    assert q.limit_number == 2

    result = (
        await Sample.find_many(Sample.increment > 2)
        .find_many(Sample.nested.optional == None)
        .sort(Sample.increment)
        .limit(2)
        .to_list()
    )  # noqa
    assert len(result) == 2
    for a in result:
        assert a.increment > 2
        assert a.nested.optional is None

    len_result = 0
    async for a in (
        Sample.find_many(Sample.increment > 2)
        .find(Sample.nested.optional == None)
        .sort(Sample.increment)
        .limit(2)
    ):  # noqa
        assert a in result
        len_result += 1

    assert len_result == len(result)


async def test_find_all(preset_documents):
    result = await Sample.find_all().to_list()
    assert len(result) == 10

    len_result = 0
    async for a in Sample.find_all():
        assert a in result
        len_result += 1

    assert len_result == len(result)


async def test_find_one(preset_documents):
    a = await Sample.find_one(Sample.integer > 1).find_one(
        Sample.nested.optional == None
    )  # noqa
    assert a.integer > 1
    assert a.nested.optional is None

    a = await Sample.find_one(Sample.integer > 100).find_one(
        Sample.nested.optional == None
    )  # noqa
    assert a is None


async def test_get(preset_documents):
    a = await Sample.find_one(Sample.integer > 1).find_one(
        Sample.nested.optional == None
    )  # noqa
    assert a.integer > 1
    assert a.nested.optional is None

    new_a = await Sample.get(a.id)
    assert new_a == a

    # check for another type
    new_a = await Sample.get(str(a.id))
    assert new_a == a


async def test_sort(preset_documents):
    q = Sample.find_many(Sample.integer > 1, sort="-integer")
    assert q.sort_expressions == [("integer", SortDirection.DESCENDING)]

    q = Sample.find_many(Sample.integer > 1, sort="integer")
    assert q.sort_expressions == [("integer", SortDirection.ASCENDING)]

    q = Sample.find_many(Sample.integer > 1).sort("-integer")
    assert q.sort_expressions == [("integer", SortDirection.DESCENDING)]

    q = (
        Sample.find_many(Sample.integer > 1)
        .find_many(Sample.integer < 100)
        .sort("-integer")
    )
    assert q.sort_expressions == [("integer", SortDirection.DESCENDING)]

    result = await Sample.find_many(
        Sample.integer > 1, sort="-integer"
    ).to_list()
    i_buf = None
    for a in result:
        if i_buf is None:
            i_buf = a.integer
        assert i_buf >= a.integer
        i_buf = a.integer

    result = await Sample.find_many(
        Sample.integer > 1, sort="+integer"
    ).to_list()
    i_buf = None
    for a in result:
        if i_buf is None:
            i_buf = a.integer
        assert i_buf <= a.integer
        i_buf = a.integer

    result = await Sample.find_many(
        Sample.integer > 1, sort="integer"
    ).to_list()
    i_buf = None
    for a in result:
        if i_buf is None:
            i_buf = a.integer
        assert i_buf <= a.integer
        i_buf = a.integer

    result = await Sample.find_many(
        Sample.integer > 1, sort=-Sample.integer
    ).to_list()
    i_buf = None
    for a in result:
        if i_buf is None:
            i_buf = a.integer
        assert i_buf >= a.integer
        i_buf = a.integer

    result = (
        await Sample.find_many(Sample.integer > 1)
        .sort([Sample.const, -Sample.integer])
        .to_list()
    )
    i_buf = None
    for a in result:
        if i_buf is None:
            i_buf = a.integer
        assert i_buf >= a.integer
        i_buf = a.integer

    with pytest.raises(TypeError):
        Sample.find_many(Sample.integer > 1, sort=1)


async def test_find_many_with_projection(preset_documents):
    class SampleProjection(BaseModel):
        string: str
        integer: int

    result = (
        await Sample.find_many(Sample.integer > 1)
        .find_many(Sample.nested.optional == None)
        .project(projection_model=SampleProjection)
        .to_list()
    )
    assert result == [
        SampleProjection(string="test_2", integer=2),
        SampleProjection(string="test_2", integer=2),
    ]

    result = (
        await Sample.find_many(Sample.integer > 1)
        .find_many(
            Sample.nested.optional == None, projection_model=SampleProjection
        )
        .to_list()
    )
    assert result == [
        SampleProjection(string="test_2", integer=2),
        SampleProjection(string="test_2", integer=2),
    ]


async def test_find_many_with_custom_projection(preset_documents):
    class SampleProjection(BaseModel):
        string: str
        i: int

        class Settings:
            projection = {"string": 1, "i": "$nested.integer"}

    result = (
        await Sample.find_many(Sample.integer > 1)
        .find_many(Sample.nested.optional == None)
        .project(projection_model=SampleProjection)
        .sort(Sample.nested.integer)
        .to_list()
    )
    assert result == [
        SampleProjection(string="test_2", i=3),
        SampleProjection(string="test_2", i=4),
    ]


async def test_find_many_with_session(preset_documents, session):
    q_1 = (
        Sample.find_many(Sample.integer > 1)
        .find_many(Sample.nested.optional == None)
        .set_session(session)
    )
    assert q_1.session == session

    q_2 = Sample.find_many(Sample.integer > 1).find_many(
        Sample.nested.optional == None, session=session
    )
    assert q_2.session == session

    result = await q_2.to_list()

    assert len(result) == 2
    for a in result:
        assert a.integer > 1
        assert a.nested.optional is None

    len_result = 0
    async for a in Sample.find_many(Sample.integer > 1).find_many(
        Sample.nested.optional == None
    ):  # noqa
        assert a in result
        len_result += 1

    assert len_result == len(result)


async def test_bson_encoders_filed_types():
    custom = DocumentWithBsonEncodersFiledsTypes(
        color="7fffd4",
        timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
    )
    c = await custom.insert()
    c_fromdb = await DocumentWithBsonEncodersFiledsTypes.find_one(
        DocumentWithBsonEncodersFiledsTypes.color == Color("7fffd4")
    )
    assert c_fromdb.color.as_hex() == c.color.as_hex()


async def test_find_by_datetime(preset_documents):
    datetime_1 = datetime.datetime.now(
        tz=datetime.timezone.utc
    ) - datetime.timedelta(days=7)
    datetime_2 = datetime.datetime.now(
        tz=datetime.timezone.utc
    ) - datetime.timedelta(days=2)
    docs = await Sample.find(
        Sample.timestamp >= datetime_1,
        Sample.timestamp <= datetime_2,
    ).to_list()
    assert len(docs) == 5


async def test_find_first_or_none(preset_documents):
    q = Sample.find(Sample.increment > 1).sort(-Sample.increment)
    doc = await q.first_or_none()
    assert doc is not None
    assert doc.increment == 9

    docs = await q.to_list()
    assert len(docs) == 8

    doc = (
        await Sample.find(Sample.increment > 9)
        .sort(-Sample.increment)
        .first_or_none()
    )
    assert doc is None


async def test_find_pymongo_kwargs(preset_documents):
    with pytest.raises(TypeError):
        await Sample.find_many(Sample.increment > 1, wrong=100).to_list()

    await Sample.find_many(
        Sample.increment > 1, Sample.integer > 1, allow_disk_use=True
    ).to_list()

    await Sample.find_many(
        Sample.increment > 1, Sample.integer > 1, hint="integer_1"
    ).to_list()

    await House.find_many(
        House.height > 1, fetch_links=True, hint="height_1"
    ).to_list()

    await House.find_many(
        House.height > 1, fetch_links=True, allowDiskUse=True
    ).to_list()

    await Sample.find_one(
        Sample.increment > 1, Sample.integer > 1, hint="integer_1"
    )

    await House.find_one(House.height > 1, fetch_links=True, hint="height_1")


def test_find_clone():
    q = (
        Sample.find_many(Sample.integer == 1)
        .find_many(Sample.nested.integer >= 2)
        .sort(Sample.integer)
        .limit(100)
    )

    new_q = q.clone()
    new_q.find(Sample.nested.integer >= 100).sort(Sample.string).limit(10)

    assert q.get_filter_query() == {
        "$and": [{"integer": 1}, {"nested.integer": {"$gte": 2}}]
    }
    assert q.sort_expressions == [("integer", SortDirection.ASCENDING)]
    assert q.limit_number == 100
    assert new_q.get_filter_query() == {
        "$and": [
            {"integer": 1},
            {"nested.integer": {"$gte": 2}},
            {"nested.integer": {"$gte": 100}},
        ]
    }
    assert new_q.sort_expressions == [
        ("integer", SortDirection.ASCENDING),
        ("string", SortDirection.ASCENDING),
    ]
    assert new_q.limit_number == 10


async def test_find_many_with_enum_in_query(preset_documents):
    class TestEnum(str, Enum):
        INTEGER = Sample.integer
        SAMPLE_NESTED_OPTIONAL = Sample.nested.optional
        CONST = "const"
        CONST_VALUE = "TEST"

    filter_query = {
        TestEnum.INTEGER: {"$gt": 1},
        TestEnum.SAMPLE_NESTED_OPTIONAL: {"$type": "null"},
        TestEnum.CONST: TestEnum.CONST_VALUE,
    }
    result = await Sample.find_many(filter_query).to_list()
    assert len(result) == 2


# @pytest.mark.asyncio
async def test_fetch_links_with_chained_delete():
    lock = await Lock(k=123).insert()
    window = await Window(x=1, y=2, lock=lock).insert()
    door = await Door(t=10, window=window, locks=[lock]).insert()

    await House(windows=[window], door=door, height=10, name="test").insert()
    await House(windows=[window], door=door, height=12, name="test2").insert()

    # Deletion with chained query and fetch_links
    deleted_count = (
        await House.find(House.height > 5, fetch_links=True)
        .find(House.height < 20)
        .delete()
    )

    assert deleted_count.deleted_count == 2

    # Confirm deletion
    remaining = await House.find_all().to_list()
    assert len(remaining) == 0



from __future__ import annotations

from types import TracebackType
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Type, Union

from pymongo import (
    DeleteMany,
    DeleteOne,
    InsertOne,
    ReplaceOne,
    UpdateMany,
    UpdateOne,
)
from pymongo.asynchronous.client_session import AsyncClientSession
from pymongo.results import BulkWriteResult

if TYPE_CHECKING:
    from beanie import Document
    from beanie.odm.union_doc import UnionDoc

_WriteOp = Union[
    InsertOne[Mapping[Any, Any]],
    DeleteOne,
    DeleteMany,
    ReplaceOne[Mapping[Any, Any]],
    UpdateOne,
    UpdateMany,
]


class BulkWriter:
    """
    A utility class for managing and executing bulk operations.

    This class facilitates the efficient execution of multiple database operations
    (e.g., inserts, updates, deletes, replacements) in a single batch. It supports asynchronous
    context management and ensures that all queued operations are committed upon exiting the context.

    Attributes:
        session Optional[AsyncClientSession]:
            The pymongo session used for transactional operations.
            Defaults to None, meaning no session is used.
        ordered Optional[bool]:
            Specifies whether operations are executed sequentially (default) or in parallel.
            - If True, operations are performed serially, stopping at the first failure.
            - If False, operations may be executed in arbitrary order, and all operations are attempted
            regardless of individual failures.
        bypass_document_validation Optional[bool]:
            If True, document-level validation is bypassed for all operations
            in the bulk write. This applies to MongoDB's schema validation rules, allowing documents that
            do not meet validation criteria to be inserted or modified. Defaults to False.
        comment Optional[Any]:
            A user-provided comment attached to the bulk operation command, useful for
            auditing and debugging purposes.
        operations List[Union[DeleteMany, DeleteOne, InsertOne, ReplaceOne, UpdateMany, UpdateOne]]:
            A list of MongoDB operations queued for bulk execution.
        object_class Type[Union[Document, UnionDoc]]:
            The document model class associated with the operations.

    Parameters:
        session Optional[AsyncClientSession]: The pymongo session for transaction support.
            Defaults to None (no session).
        ordered Optional[bool]: Specifies whether operations are executed in sequence (True) or in parallel (False).
            Defaults to True.
        bypass_document_validation Optional[bool]: Allows the bulk operation to bypass document-level validation.
            This is particularly useful when working with schemas that are being phased in or for bulk imports
            where strict validation may not be necessary. Defaults to False.
        comment Optional[Any]: A custom comment attached to the bulk operation.
            Defaults to None.
        object_class Type[Union[Document, UnionDoc]]: The document model class associated with the operations.
    """

    def __init__(
        self,
        session: Optional[AsyncClientSession] = None,
        ordered: bool = True,
        object_class: Optional[Type[Union[Document, UnionDoc]]] = None,
        bypass_document_validation: Optional[bool] = False,
        comment: Optional[Any] = None,
    ) -> None:
        self.operations: List[_WriteOp] = []
        self.session = session
        self.ordered = ordered
        self.object_class = object_class
        self.bypass_document_validation = bypass_document_validation
        self.comment = comment
        self._collection_name: Optional[str] = (
            object_class.get_collection_name() if object_class else None
        )

    async def __aenter__(self) -> "BulkWriter":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        if exc_type is None:
            await self.commit()

    async def commit(self) -> Optional[BulkWriteResult]:
        """
        Commit all queued operations to the database.

        Executes all queued operations in a single bulk write request. If there
        are no operations to commit, it returns ``None``.

        :return: The result of the bulk write operation if operations are committed.
                Returns ``None`` if there are no operations to execute.
        :rtype: Optional[BulkWriteResult]

        :raises ValueError:
            If the object_class is not specified before committing.
        """
        if not self.operations:
            return None
        if not self.object_class:
            raise ValueError(
                "The document model class must be specified before committing operations."
            )
        return await self.object_class.get_pymongo_collection().bulk_write(
            self.operations,
            ordered=self.ordered,
            bypass_document_validation=self.bypass_document_validation,
            session=self.session,
            comment=self.comment,
        )

    def add_operation(
        self,
        object_class: Type[Union[Document, UnionDoc]],
        operation: _WriteOp,
    ):
        """
        Add an operation to the queue.

        This method adds a MongoDB operation to the BulkWriter's operation queue.
        All operations in the queue must belong to the same collection.

        :param object_class: Type[Union[Document, UnionDoc]]
            The document model class associated with the operation.
        :param operation: Union[DeleteMany, DeleteOne, InsertOne, ReplaceOne, UpdateMany, UpdateOne]
            The MongoDB operation to add to the queue.
        :raises ValueError:
            If the collection differs from the one already associated with the BulkWriter.
        """
        if self.object_class is None:
            self.object_class = object_class
            self._collection_name = object_class.get_collection_name()
        else:
            if object_class.get_collection_name() != self._collection_name:
                raise ValueError(
                    "All the operations should be for a same collection name"
                )
        self.operations.append(operation)



import pytest

from beanie.odm.utils.dump import get_dict
from beanie.odm.utils.parsing import parse_obj
from tests.odm.models import SampleLazyParsing


@pytest.fixture
async def docs():
    for i in range(10):
        await SampleLazyParsing(i=i, s=str(i)).insert()


class TestLazyParsing:
    async def test_find_all(self, docs):
        found_docs = await SampleLazyParsing.all(lazy_parse=True).to_list()
        saved_state = found_docs[0].get_saved_state()
        assert "_id" in saved_state
        del saved_state["_id"]
        assert found_docs[0].get_saved_state() == {}
        assert found_docs[0].i == 0
        assert found_docs[0].s == "0"
        assert found_docs[1]._store["i"] == 1
        assert found_docs[1]._store["s"] == "1"
        assert get_dict(found_docs[2])["i"] == 2
        assert get_dict(found_docs[2])["s"] == "2"

    async def test_find_many(self, docs):
        found_docs = await SampleLazyParsing.find(
            SampleLazyParsing.i <= 5, lazy_parse=True
        ).to_list()
        saved_state = found_docs[0].get_saved_state()
        assert "_id" in saved_state
        del saved_state["_id"]
        assert found_docs[0].get_saved_state() == {}
        assert found_docs[0].i == 0
        assert found_docs[0].s == "0"
        assert found_docs[1]._store["i"] == 1
        assert found_docs[1]._store["s"] == "1"
        assert get_dict(found_docs[2])["i"] == 2
        assert get_dict(found_docs[2])["s"] == "2"

    async def test_save_changes(self, docs):
        found_docs = await SampleLazyParsing.all(lazy_parse=True).to_list()
        doc = found_docs[0]
        doc.i = 1000
        await doc.save_changes()
        new_doc = await SampleLazyParsing.find_one(SampleLazyParsing.s == "0")
        assert new_doc.i == 1000

    async def test_default_list(self):
        res = parse_obj(SampleLazyParsing, {"i": "1", "s": "1"})
        assert res.lst == []



from tests.odm.models import DocumentTestModel


async def test_distinct_unique(documents, document_not_inserted):
    await documents(1, "uno")
    await documents(2, "dos")
    await documents(3, "cuatro")
    expected_result = ["cuatro", "dos", "uno"]
    unique_test_strs = await DocumentTestModel.distinct("test_str", {})
    assert unique_test_strs == expected_result
    document_not_inserted.test_str = "uno"
    await document_not_inserted.insert()
    another_unique_test_strs = await DocumentTestModel.distinct("test_str", {})
    assert another_unique_test_strs == expected_result


async def test_distinct_different_value(documents, document_not_inserted):
    await documents(1, "uno")
    await documents(2, "dos")
    await documents(3, "cuatro")
    expected_result = ["cuatro", "dos", "uno"]
    unique_test_strs = await DocumentTestModel.distinct("test_str", {})
    assert unique_test_strs == expected_result
    document_not_inserted.test_str = "diff_val"
    await document_not_inserted.insert()
    another_unique_test_strs = await DocumentTestModel.distinct("test_str", {})
    assert not another_unique_test_strs == expected_result
    assert another_unique_test_strs == ["cuatro", "diff_val", "dos", "uno"]



from uuid import uuid4

import pytest

from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import (
    DocumentWithBackLink,
    DocumentWithDecimalField,
    DocumentWithIndexedObjectId,
    DocumentWithLink,
    DocumentWithListBackLink,
    DocumentWithListLink,
    DocumentWithOptionalLink,
)


def test_schema_export_of_model_with_decimal_field():
    doc = DocumentWithDecimalField(amt=0.1, other_amt=3.5)
    if IS_PYDANTIC_V2:
        json_schema = doc.model_json_schema()

        assert json_schema["properties"]["amt"]["anyOf"][0]["type"] == "number"
        assert json_schema["properties"]["amt"]["anyOf"][1]["type"] == "string"
        assert (
            json_schema["properties"]["other_amt"]["anyOf"][0]["type"]
            == "number"
        )
        assert (
            json_schema["properties"]["other_amt"]["anyOf"][1]["type"]
            == "string"
        )

    else:
        json_schema = doc.schema()

        assert json_schema["properties"]["amt"]["type"] == "number"
        assert json_schema["properties"]["other_amt"]["type"] == "number"


def test_schema_export_of_model_with_pydanticobjectid():
    doc = DocumentWithIndexedObjectId(
        pyid="5f8d0a8b0b7e3a1e4c9f4b1e", uuid=uuid4(), email="test@test.com"
    )
    if IS_PYDANTIC_V2:
        json_schema = doc.model_json_schema()

        assert (
            json_schema["properties"]["_id"]["anyOf"][0]["$ref"]
            == "#/$defs/PydanticObjectId"
        )
        assert (
            json_schema["properties"]["pyid"]["$ref"]
            == "#/$defs/PydanticObjectId"
        )
    else:
        json_schema = doc.schema()

        assert json_schema["properties"]["_id"]["type"] == "string"
        assert json_schema["properties"]["pyid"]["type"] == "string"


def test_schema_export_of_model_with_link():
    if IS_PYDANTIC_V2:
        json_schema = DocumentWithLink.model_json_schema()

        link_alternate_representation = json_schema["properties"]["link"][
            "anyOf"
        ]
    else:
        json_schema = DocumentWithLink.schema()

        link_alternate_representation = json_schema["definitions"][
            "DocumentWithLink"
        ]["properties"]["link"]["anyOf"]

    assert link_alternate_representation[0]["type"] == "object"
    assert link_alternate_representation[1]["type"] == "object"


@pytest.mark.skipif(
    not IS_PYDANTIC_V2,
    reason="schema dumping support is more complete with pydantic v2",
)
def test_schema_export_of_model_with_optional_link():
    if IS_PYDANTIC_V2:
        json_schema = DocumentWithOptionalLink.model_json_schema()
    else:
        json_schema = DocumentWithOptionalLink.schema()

    link_alternate_representation = json_schema["properties"]["link"]["anyOf"]
    assert link_alternate_representation[0]["type"] == "object"
    assert link_alternate_representation[1]["type"] == "object"
    assert link_alternate_representation[2]["type"] == "null"


def test_schema_export_of_model_with_list_link():
    if IS_PYDANTIC_V2:
        json_schema = DocumentWithListLink.model_json_schema()

        link_alternate_representation = json_schema["properties"]["link"][
            "items"
        ]["anyOf"]
        link_definition = json_schema["properties"]["link"]["type"]
    else:
        json_schema = DocumentWithListLink.schema()

        link_alternate_representation = json_schema["definitions"][
            "DocumentWithListLink"
        ]["properties"]["link"]["items"]["anyOf"]
        link_definition = json_schema["definitions"]["DocumentWithListLink"][
            "properties"
        ]["link"]["type"]

    assert link_definition == "array"
    assert link_alternate_representation[0]["type"] == "object"
    assert link_alternate_representation[1]["type"] == "object"


def test_schema_export_of_model_with_backlink():
    if IS_PYDANTIC_V2:
        json_schema = DocumentWithBackLink.model_json_schema()

        back_link_definition = json_schema["properties"]["back_link"]["type"]
    else:
        json_schema = DocumentWithBackLink.schema()

        back_link_definition = json_schema["definitions"][
            "DocumentWithBackLink"
        ]["properties"]["back_link"]["anyOf"][1]["type"]

    assert back_link_definition == "object"


def test_schema_export_of_model_with_list_backlink():
    if IS_PYDANTIC_V2:
        json_schema = DocumentWithListBackLink.model_json_schema()

        assert json_schema["properties"]["back_link"]["type"] == "array"
        assert (
            json_schema["properties"]["back_link"]["items"]["type"] == "object"
        )
    else:
        json_schema = DocumentWithListBackLink.schema()

        assert (
            json_schema["definitions"]["DocumentWithListBackLink"][
                "properties"
            ]["back_link"]["type"]
            == "array"
        )
        assert (
            json_schema["definitions"]["DocumentWithListBackLink"][
                "properties"
            ]["back_link"]["items"]["anyOf"][1]["type"]
            == "object"
        )


@pytest.mark.skipif(
    not IS_PYDANTIC_V2,
    reason="schema dumping support is more complete with pydantic v2",
)
@pytest.mark.parametrize(
    "model_type",
    [
        DocumentWithDecimalField,
        DocumentWithIndexedObjectId,
        DocumentWithLink,
        DocumentWithOptionalLink,
        DocumentWithListLink,
        DocumentWithBackLink,
        DocumentWithListBackLink,
    ],
)
def test_json_serialization_of_model(model_type):
    validation_schema = model_type.model_json_schema(mode="serialization")
    assert validation_schema is not None
    assert isinstance(validation_schema, dict)



import pytest
from pydantic import BaseModel, Field

from beanie import Link, PydanticObjectId
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import DocumentTestModelWithSoftDelete


class TestModel(BaseModel):
    my_id: PydanticObjectId = Field(default_factory=PydanticObjectId)
    fake_doc: Link[DocumentTestModelWithSoftDelete]


def data_maker():
    return TestModel(
        my_id="5f4e3f3b7c0c9d001f7d4c8e",
        fake_doc=DocumentTestModelWithSoftDelete(
            test_int=1, test_str="test", id="5f4e3f3b7c0c9d001f7d4c8f"
        ),
    )


@pytest.mark.skipif(
    not IS_PYDANTIC_V2,
    reason="model dumping support is more complete with pydantic v2",
)
def test_id_types_preserved_when_dumping_to_python():
    dumped = data_maker().model_dump(mode="python")
    assert isinstance(dumped["my_id"], PydanticObjectId)
    assert isinstance(dumped["fake_doc"]["id"], PydanticObjectId)


@pytest.mark.skipif(
    not IS_PYDANTIC_V2,
    reason="model dumping support is more complete with pydantic v2",
)
def test_id_types_serialized_when_dumping_to_json():
    dumped = data_maker().model_dump(mode="json")
    assert isinstance(dumped["my_id"], str)
    assert isinstance(dumped["fake_doc"]["id"], str)



import pytest

from beanie import init_beanie
from beanie.migrations.models import MigrationLog


@pytest.fixture(autouse=True)
async def init(db):
    await init_beanie(
        database=db,
        document_models=[
            MigrationLog,
        ],
    )


@pytest.fixture(autouse=True)
async def remove_migrations_log(db, init):
    await MigrationLog.delete_all()



import pytest
from pymongo.errors import BulkWriteError

from beanie.odm.bulk import BulkWriter
from beanie.odm.operators.update.general import Set
from tests.odm.models import (
    DocumentMultiModelOne,
    DocumentMultiModelTwo,
    DocumentTestModel,
    DocumentUnion,
    SubDocument,
)


async def test_insert(documents_not_inserted):
    documents = documents_not_inserted(2)
    async with BulkWriter() as bulk_writer:
        await DocumentTestModel.insert_one(
            documents[0], bulk_writer=bulk_writer
        )
        await DocumentTestModel.insert_one(
            documents[1], bulk_writer=bulk_writer
        )

    new_documents = await DocumentTestModel.find_all().to_list()
    assert len(new_documents) == 2


async def test_update(documents):
    await documents(5)
    doc = await DocumentTestModel.find_one(DocumentTestModel.test_int == 0)
    doc.test_int = 100
    async with BulkWriter() as bulk_writer:
        await doc.save_changes(bulk_writer=bulk_writer)
        await DocumentTestModel.find_one(
            DocumentTestModel.test_int == 1
        ).update(
            Set({DocumentTestModel.test_int: 1000}), bulk_writer=bulk_writer
        )
        await DocumentTestModel.find(DocumentTestModel.test_int < 100).update(
            Set({DocumentTestModel.test_int: 2000}), bulk_writer=bulk_writer
        )

    assert len(await DocumentTestModel.find_all().to_list()) == 5
    assert (
        len(
            await DocumentTestModel.find(
                DocumentTestModel.test_int == 100
            ).to_list()
        )
        == 1
    )
    assert (
        len(
            await DocumentTestModel.find(
                DocumentTestModel.test_int == 1000
            ).to_list()
        )
        == 1
    )
    assert (
        len(
            await DocumentTestModel.find(
                DocumentTestModel.test_int == 2000
            ).to_list()
        )
        == 3
    )


async def test_unordered_update(documents, document):
    await documents(5)
    doc = await DocumentTestModel.find_one(DocumentTestModel.test_int == 0)
    doc.test_int = 100
    with pytest.raises(BulkWriteError):
        async with BulkWriter(ordered=False) as bulk_writer:
            await DocumentTestModel.insert_one(
                document, bulk_writer=bulk_writer
            )
            await doc.save_changes(bulk_writer=bulk_writer)

    assert len(await DocumentTestModel.find_all().to_list()) == 6
    assert (
        len(
            await DocumentTestModel.find(
                DocumentTestModel.test_int == 100
            ).to_list()
        )
        == 1
    )


async def test_delete(documents):
    await documents(5)
    doc = await DocumentTestModel.find_one(DocumentTestModel.test_int == 0)
    async with BulkWriter() as bulk_writer:
        await doc.delete(bulk_writer=bulk_writer)
        await DocumentTestModel.find_one(
            DocumentTestModel.test_int == 1
        ).delete(bulk_writer=bulk_writer)
        await DocumentTestModel.find(DocumentTestModel.test_int < 4).delete(
            bulk_writer=bulk_writer
        )

    assert len(await DocumentTestModel.find_all().to_list()) == 1


async def test_replace(documents, document_not_inserted):
    await documents(5)
    doc = await DocumentTestModel.find_one(DocumentTestModel.test_int == 0)
    doc.test_int = 100
    async with BulkWriter() as bulk_writer:
        await doc.replace(bulk_writer=bulk_writer)

        document_not_inserted.test_int = 100

        await DocumentTestModel.find_one(
            DocumentTestModel.test_int == 1
        ).replace_one(document_not_inserted, bulk_writer=bulk_writer)

    assert len(await DocumentTestModel.find_all().to_list()) == 5
    assert (
        len(
            await DocumentTestModel.find(
                DocumentTestModel.test_int == 100
            ).to_list()
        )
        == 2
    )


async def test_internal_error(document):
    with pytest.raises(BulkWriteError):
        async with BulkWriter() as bulk_writer:
            await DocumentTestModel.insert_one(
                document, bulk_writer=bulk_writer
            )


async def test_native_upsert_found(documents, document_not_inserted):
    await documents(5)
    document_not_inserted.test_int = -1000
    async with BulkWriter() as bulk_writer:
        await DocumentTestModel.find_one(
            DocumentTestModel.test_int == 1
        ).update_one(
            {
                "$addToSet": {
                    "test_list": {
                        "$each": [
                            SubDocument(test_str="TEST_ONE"),
                            SubDocument(test_str="TEST_TWO"),
                        ]
                    }
                },
                "$setOnInsert": {},
            },
            bulk_writer=bulk_writer,
            upsert=True,
        )
        await bulk_writer.commit()

    doc = await DocumentTestModel.find_one(DocumentTestModel.test_int == 1)
    assert len(doc.test_list) == 4


async def test_native_upsert_not_found(documents, document_not_inserted):
    await documents(5)
    document_not_inserted.test_int = -1000
    async with BulkWriter() as bulk_writer:
        await DocumentTestModel.find_one(
            DocumentTestModel.test_int == -1000
        ).update_one(
            {
                "$addToSet": {
                    "test_list": {
                        "$each": [
                            SubDocument(test_str="TEST_ONE"),
                            SubDocument(test_str="TEST_TWO"),
                        ]
                    }
                },
                "$setOnInsert": {"TEST": "VALUE"},
            },
            bulk_writer=bulk_writer,
            upsert=True,
        )
        await bulk_writer.commit()

    assert await DocumentTestModel.count() == 6


async def test_different_models_same_collection():
    async with BulkWriter() as bulk_writer:
        await DocumentMultiModelOne.insert_one(
            DocumentMultiModelOne(), bulk_writer=bulk_writer
        )
        await DocumentMultiModelTwo.insert_one(
            DocumentMultiModelTwo(), bulk_writer=bulk_writer
        )
    assert len(await DocumentUnion.find(with_children=True).to_list()) == 2


async def test_empty_operations():
    bulk = BulkWriter()
    await DocumentMultiModelOne.insert_one(
        DocumentMultiModelOne(), bulk_writer=bulk
    )
    await DocumentMultiModelOne.insert_one(
        DocumentMultiModelOne(), bulk_writer=bulk
    )
    assert len(bulk.operations) == 2
    bulk.operations = []
    bulk_result = await bulk.commit()
    assert bulk_result == None
    assert len(await DocumentMultiModelOne.find().to_list()) == 0


async def test_ordered_bulk(documents):
    await documents(1)
    doc = await DocumentMultiModelOne.insert_one(DocumentMultiModelOne())
    assert doc
    assert doc.id
    with pytest.raises(BulkWriteError):
        async with BulkWriter(ordered=True) as bulk_writer:
            doc1 = DocumentMultiModelOne()
            doc1.id = doc.id
            await DocumentMultiModelOne.insert_one(
                doc1, bulk_writer=bulk_writer
            )
            await DocumentMultiModelOne.insert_one(
                DocumentMultiModelOne(), bulk_writer=bulk_writer
            )

    assert len(await DocumentMultiModelOne.find_all().to_list()) == 1


async def test_bulk_writer():
    assert isinstance(DocumentMultiModelOne.bulk_writer(), BulkWriter)
    assert isinstance(DocumentUnion.bulk_writer(), BulkWriter)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

from beanie.odm.fields import LinkInfo, LinkTypes

if TYPE_CHECKING:
    from beanie import Document


# TODO: check if this is the most efficient way for
#  appending subqueries to the queries var


def construct_lookup_queries(
    cls: Type["Document"],
    nesting_depth: Optional[int] = None,
    nesting_depths_per_field: Optional[Dict[str, int]] = None,
) -> List[Dict[str, Any]]:
    queries: List = []
    link_fields = cls.get_link_fields()
    if link_fields is not None:
        for link_info in link_fields.values():
            final_nesting_depth = (
                nesting_depths_per_field.get(link_info.field_name, None)
                if nesting_depths_per_field is not None
                else None
            )
            if final_nesting_depth is None:
                final_nesting_depth = nesting_depth
            construct_query(
                link_info=link_info,
                queries=queries,
                database_major_version=cls._database_major_version,
                current_depth=final_nesting_depth,
            )
    return queries


def construct_query(
    link_info: LinkInfo,
    queries: List,
    database_major_version: int,
    current_depth: Optional[int] = None,
):
    if link_info.is_fetchable is False or (
        current_depth is not None and current_depth <= 0
    ):
        return
    if link_info.link_type in [
        LinkTypes.DIRECT,
        LinkTypes.OPTIONAL_DIRECT,
    ]:
        if database_major_version >= 5 or link_info.nested_links is None:
            lookup_steps = [
                {
                    "$lookup": {
                        "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                        "localField": f"{link_info.lookup_field_name}.$id",
                        "foreignField": "_id",
                        "as": f"_link_{link_info.field_name}",
                    }
                },
                {
                    "$unwind": {
                        "path": f"$_link_{link_info.field_name}",
                        "preserveNullAndEmptyArrays": True,
                    }
                },
                {
                    "$addFields": {
                        link_info.field_name: {
                            "$cond": {
                                "if": {
                                    "$ifNull": [
                                        f"$_link_{link_info.field_name}",
                                        False,
                                    ]
                                },
                                "then": f"$_link_{link_info.field_name}",
                                "else": f"${link_info.field_name}",
                            }
                        }
                    }
                },
                {"$project": {f"_link_{link_info.field_name}": 0}},
            ]  # type: ignore
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            if link_info.nested_links is not None:
                lookup_steps[0]["$lookup"]["pipeline"] = []  # type: ignore
                for nested_link in link_info.nested_links:
                    construct_query(
                        link_info=link_info.nested_links[nested_link],
                        queries=lookup_steps[0]["$lookup"]["pipeline"],  # type: ignore
                        database_major_version=database_major_version,
                        current_depth=new_depth,
                    )
            queries += lookup_steps

        else:
            lookup_steps = [
                {
                    "$lookup": {
                        "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                        "let": {
                            "link_id": f"${link_info.lookup_field_name}.$id"
                        },
                        "as": f"_link_{link_info.field_name}",
                        "pipeline": [
                            {
                                "$match": {
                                    "$expr": {"$eq": ["$_id", "$$link_id"]}
                                }
                            },
                        ],
                    }
                },
                {
                    "$unwind": {
                        "path": f"$_link_{link_info.field_name}",
                        "preserveNullAndEmptyArrays": True,
                    }
                },
                {
                    "$addFields": {
                        link_info.field_name: {
                            "$cond": {
                                "if": {
                                    "$ifNull": [
                                        f"$_link_{link_info.field_name}",
                                        False,
                                    ]
                                },
                                "then": f"$_link_{link_info.field_name}",
                                "else": f"${link_info.field_name}",
                            }
                        }
                    }
                },
                {"$project": {f"_link_{link_info.field_name}": 0}},
            ]
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            for nested_link in link_info.nested_links:
                construct_query(
                    link_info=link_info.nested_links[nested_link],
                    queries=lookup_steps[0]["$lookup"]["pipeline"],  # type: ignore
                    database_major_version=database_major_version,
                    current_depth=new_depth,
                )
            queries += lookup_steps

    elif link_info.link_type in [
        LinkTypes.BACK_DIRECT,
        LinkTypes.OPTIONAL_BACK_DIRECT,
    ]:
        if database_major_version >= 5 or link_info.nested_links is None:
            lookup_steps = [
                {
                    "$lookup": {
                        "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                        "localField": "_id",
                        "foreignField": f"{link_info.lookup_field_name}.$id",
                        "as": f"_link_{link_info.field_name}",
                    }
                },
                {
                    "$unwind": {
                        "path": f"$_link_{link_info.field_name}",
                        "preserveNullAndEmptyArrays": True,
                    }
                },
                {
                    "$addFields": {
                        link_info.field_name: {
                            "$cond": {
                                "if": {
                                    "$ifNull": [
                                        f"$_link_{link_info.field_name}",
                                        False,
                                    ]
                                },
                                "then": f"$_link_{link_info.field_name}",
                                "else": f"${link_info.field_name}",
                            }
                        }
                    }
                },
                {"$project": {f"_link_{link_info.field_name}": 0}},
            ]  # type: ignore
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            if link_info.nested_links is not None:
                lookup_steps[0]["$lookup"]["pipeline"] = []  # type: ignore
                for nested_link in link_info.nested_links:
                    construct_query(
                        link_info=link_info.nested_links[nested_link],
                        queries=lookup_steps[0]["$lookup"]["pipeline"],  # type: ignore
                        database_major_version=database_major_version,
                        current_depth=new_depth,
                    )
            queries += lookup_steps

        else:
            lookup_steps = [
                {
                    "$lookup": {
                        "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                        "let": {"link_id": "$_id"},
                        "as": f"_link_{link_info.field_name}",
                        "pipeline": [
                            {
                                "$match": {
                                    "$expr": {
                                        "$eq": [
                                            f"${link_info.lookup_field_name}.$id",
                                            "$$link_id",
                                        ]
                                    }
                                }
                            },
                        ],
                    }
                },
                {
                    "$unwind": {
                        "path": f"$_link_{link_info.field_name}",
                        "preserveNullAndEmptyArrays": True,
                    }
                },
                {
                    "$addFields": {
                        link_info.field_name: {
                            "$cond": {
                                "if": {
                                    "$ifNull": [
                                        f"$_link_{link_info.field_name}",
                                        False,
                                    ]
                                },
                                "then": f"$_link_{link_info.field_name}",
                                "else": f"${link_info.field_name}",
                            }
                        }
                    }
                },
                {"$project": {f"_link_{link_info.field_name}": 0}},
            ]
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            for nested_link in link_info.nested_links:
                construct_query(
                    link_info=link_info.nested_links[nested_link],
                    queries=lookup_steps[0]["$lookup"]["pipeline"],  # type: ignore
                    database_major_version=database_major_version,
                    current_depth=new_depth,
                )
            queries += lookup_steps

    elif link_info.link_type in [
        LinkTypes.LIST,
        LinkTypes.OPTIONAL_LIST,
    ]:
        if database_major_version >= 5 or link_info.nested_links is None:
            queries.append(
                {
                    "$lookup": {
                        "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                        "localField": f"{link_info.lookup_field_name}.$id",
                        "foreignField": "_id",
                        "as": link_info.field_name,
                    }
                }
            )
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            if link_info.nested_links is not None:
                queries[-1]["$lookup"]["pipeline"] = []
                for nested_link in link_info.nested_links:
                    construct_query(
                        link_info=link_info.nested_links[nested_link],
                        queries=queries[-1]["$lookup"]["pipeline"],
                        database_major_version=database_major_version,
                        current_depth=new_depth,
                    )
        else:
            lookup_step = {
                "$lookup": {
                    "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                    "let": {"link_id": f"${link_info.lookup_field_name}.$id"},
                    "as": link_info.field_name,
                    "pipeline": [
                        {"$match": {"$expr": {"$in": ["$_id", "$$link_id"]}}},
                    ],
                }
            }
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            for nested_link in link_info.nested_links:
                construct_query(
                    link_info=link_info.nested_links[nested_link],
                    queries=lookup_step["$lookup"]["pipeline"],
                    database_major_version=database_major_version,
                    current_depth=new_depth,
                )
            queries.append(lookup_step)

    elif link_info.link_type in [
        LinkTypes.BACK_LIST,
        LinkTypes.OPTIONAL_BACK_LIST,
    ]:
        if database_major_version >= 5 or link_info.nested_links is None:
            queries.append(
                {
                    "$lookup": {
                        "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                        "localField": "_id",
                        "foreignField": f"{link_info.lookup_field_name}.$id",
                        "as": link_info.field_name,
                    }
                }
            )
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            if link_info.nested_links is not None:
                queries[-1]["$lookup"]["pipeline"] = []
                for nested_link in link_info.nested_links:
                    construct_query(
                        link_info=link_info.nested_links[nested_link],
                        queries=queries[-1]["$lookup"]["pipeline"],
                        database_major_version=database_major_version,
                        current_depth=new_depth,
                    )
        else:
            lookup_step = {
                "$lookup": {
                    "from": link_info.document_class.get_pymongo_collection().name,  # type: ignore
                    "let": {"link_id": "$_id"},
                    "as": link_info.field_name,
                    "pipeline": [
                        {
                            "$match": {
                                "$expr": {
                                    "$in": [
                                        "$$link_id",
                                        f"${link_info.lookup_field_name}.$id",
                                    ]
                                }
                            }
                        }
                    ],
                }
            }
            new_depth = (
                current_depth - 1 if current_depth is not None else None
            )
            for nested_link in link_info.nested_links:
                construct_query(
                    link_info=link_info.nested_links[nested_link],
                    queries=lookup_step["$lookup"]["pipeline"],
                    database_major_version=database_major_version,
                    current_depth=new_depth,
                )
            queries.append(lookup_step)

    return queries


def split_text_query(
    query: Dict[str, Any],
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """Divide query into text and non-text matches

    :param query: Dict[str, Any] - query dict
    :return: Tuple[Dict[str, Any], Dict[str, Any]] - text and non-text queries,
        respectively
    """

    root_text_query_args: Dict[str, Any] = query.get("$text", None)
    root_non_text_queries: Dict[str, Any] = {
        k: v for k, v in query.items() if k not in {"$text", "$and"}
    }

    text_queries: List[Dict[str, Any]] = (
        [{"$text": root_text_query_args}] if root_text_query_args else []
    )
    non_text_queries: List[Dict[str, Any]] = (
        [root_non_text_queries] if root_non_text_queries else []
    )

    for match_case in query.get("$and", []):
        if "$text" in match_case:
            text_queries.append(match_case)
        else:
            non_text_queries.append(match_case)

    return text_queries, non_text_queries



from typing import Dict, ForwardRef, Type, Union

from pydantic import BaseModel


class DocsRegistry:
    _registry: Dict[str, Type[BaseModel]] = {}

    @classmethod
    def register(cls, name: str, doc_type: Type[BaseModel]):
        cls._registry[name] = doc_type

    @classmethod
    def get(cls, name: str) -> Type[BaseModel]:
        return cls._registry[name]

    @classmethod
    def evaluate_fr(cls, forward_ref: Union[ForwardRef, Type]):
        """
        Evaluate forward ref

        :param forward_ref: ForwardRef - forward ref to evaluate
        :return: Type[BaseModel] - class of the forward ref
        """
        if (
            isinstance(forward_ref, ForwardRef)
            and forward_ref.__forward_arg__ in cls._registry
        ):
            return cls._registry[forward_ref.__forward_arg__]
        else:
            return forward_ref



import decimal

import bson
import pydantic
from typing_extensions import Annotated

DecimalAnnotation = Annotated[
    decimal.Decimal,
    pydantic.BeforeValidator(
        lambda v: v.to_decimal() if isinstance(v, bson.Decimal128) else v
    ),
]



from tests.odm.models import DocumentTestModelWithSoftDelete


async def test_get_item(document_soft_delete_not_inserted):
    # insert a document with soft delete
    result = await document_soft_delete_not_inserted.insert()

    # get from db by id
    document = await DocumentTestModelWithSoftDelete.get(document_id=result.id)

    assert document.is_deleted() is False
    assert document.deleted_at is None
    assert document.test_int == result.test_int
    assert document.test_str == result.test_str

    # # delete the document
    await document.delete()
    assert document.is_deleted() is True

    # check if document exist with `.get()`
    document = await DocumentTestModelWithSoftDelete.get(document_id=result.id)
    assert document is None

    # check document exist in trashed
    results = (
        await DocumentTestModelWithSoftDelete.find_many_in_all().to_list()
    )
    assert len(results) == 1


async def test_find_one(document_soft_delete_not_inserted):
    result = await document_soft_delete_not_inserted.insert()

    # # delete the document
    await result.delete()

    # check if document exist with `.find_one()`
    document = await DocumentTestModelWithSoftDelete.find_one(
        DocumentTestModelWithSoftDelete.id == result.id
    )
    assert document is None


async def test_find(documents_soft_delete_not_inserted):
    # insert 3 documents
    inserted_docs = []
    for doc in documents_soft_delete_not_inserted:
        result = await doc.insert()
        inserted_docs.append(result)

    # use `.find_many()` to get them all
    results = await DocumentTestModelWithSoftDelete.find().to_list()
    assert len(results) == 3

    # delete one of them
    await inserted_docs[0].delete()

    # check items in with `.find_many()`
    results = await DocumentTestModelWithSoftDelete.find_many().to_list()

    assert len(results) == 2

    founded_documents_id = [doc.id for doc in results]
    assert inserted_docs[0].id not in founded_documents_id

    # check in trashed items
    results = (
        await DocumentTestModelWithSoftDelete.find_many_in_all().to_list()
    )
    assert len(results) == 3


async def test_find_many(documents_soft_delete_not_inserted):
    # insert 2 documents
    item_1 = await documents_soft_delete_not_inserted[0].insert()
    item_2 = await documents_soft_delete_not_inserted[1].insert()

    # use `.find_many()` to get them all
    results = await DocumentTestModelWithSoftDelete.find_many().to_list()
    assert len(results) == 2

    # delete one of them
    await item_1.delete()

    # check items in with `.find_many()`
    results = await DocumentTestModelWithSoftDelete.find_many().to_list()

    assert len(results) == 1
    assert results[0].id == item_2.id

    # check in trashed items
    results = (
        await DocumentTestModelWithSoftDelete.find_many_in_all().to_list()
    )
    assert len(results) == 2


async def test_hard_delete(document_soft_delete_not_inserted):
    result = await document_soft_delete_not_inserted.insert()
    await result.hard_delete()

    # check items in with `.find_many()`
    results = await DocumentTestModelWithSoftDelete.find_many().to_list()
    assert len(results) == 0

    # check in trashed
    results = (
        await DocumentTestModelWithSoftDelete.find_many_in_all().to_list()
    )
    assert len(results) == 0



import dataclasses as dc
import datetime
import decimal
import enum
import ipaddress
import operator
import pathlib
import re
import uuid
from enum import Enum
from typing import (
    Any,
    Callable,
    Container,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
)

import bson
import pydantic

import beanie
from beanie.odm.fields import Link, LinkTypes
from beanie.odm.utils.pydantic import (
    IS_PYDANTIC_V2,
    IS_PYDANTIC_V2_10,
    get_model_fields,
)

SingleArgCallable = Callable[[Any], Any]
DEFAULT_CUSTOM_ENCODERS: MutableMapping[type, SingleArgCallable] = {
    ipaddress.IPv4Address: str,
    ipaddress.IPv4Interface: str,
    ipaddress.IPv4Network: str,
    ipaddress.IPv6Address: str,
    ipaddress.IPv6Interface: str,
    ipaddress.IPv6Network: str,
    pathlib.PurePath: str,
    pydantic.SecretBytes: pydantic.SecretBytes.get_secret_value,
    pydantic.SecretStr: pydantic.SecretStr.get_secret_value,
    datetime.date: lambda d: datetime.datetime.combine(d, datetime.time.min),
    datetime.timedelta: operator.methodcaller("total_seconds"),
    enum.Enum: operator.attrgetter("value"),
    Link: operator.attrgetter("ref"),
    bytes: bson.Binary,
    decimal.Decimal: bson.Decimal128,
    uuid.UUID: bson.Binary.from_uuid,
    re.Pattern: bson.Regex.from_native,
}
if IS_PYDANTIC_V2:
    from pydantic_core import Url

    DEFAULT_CUSTOM_ENCODERS[Url] = str

if IS_PYDANTIC_V2_10:
    from pydantic import AnyUrl

    DEFAULT_CUSTOM_ENCODERS[AnyUrl] = str

BSON_SCALAR_TYPES = (
    type(None),
    str,
    int,
    float,
    datetime.datetime,
    bson.Binary,
    bson.DBRef,
    bson.Decimal128,
    bson.MaxKey,
    bson.MinKey,
    bson.ObjectId,
    bson.Regex,
)


@dc.dataclass
class Encoder:
    """
    BSON encoding class
    """

    exclude: Container[str] = frozenset()
    include: Container[str] = frozenset()
    custom_encoders: Mapping[type, SingleArgCallable] = dc.field(
        default_factory=dict
    )
    to_db: bool = False
    keep_nulls: bool = True

    def _encode_document(self, obj: "beanie.Document") -> Mapping[str, Any]:
        obj.parse_store()
        settings = obj.get_settings()
        obj_dict = {}
        if settings.union_doc is not None:
            obj_dict[settings.class_id] = (
                settings.union_doc_alias or obj.__class__.__name__
            )
        if obj._class_id:
            obj_dict[settings.class_id] = obj._class_id

        link_fields = obj.get_link_fields() or {}
        sub_encoder = Encoder(
            # don't propagate self.exclude to subdocuments
            custom_encoders=settings.bson_encoders,
            to_db=self.to_db,
            keep_nulls=self.keep_nulls,
        )
        for key, value in self._iter_model_items(obj):
            if key in link_fields:
                link_type = link_fields[key].link_type
                if link_type in (LinkTypes.DIRECT, LinkTypes.OPTIONAL_DIRECT):
                    if value is not None:
                        value = value.to_ref()
                elif link_type in (LinkTypes.LIST, LinkTypes.OPTIONAL_LIST):
                    if value is not None:
                        value = [link.to_ref() for link in value]
                elif self.to_db:
                    continue
            obj_dict[key] = sub_encoder.encode(value)
        return obj_dict

    def encode(self, obj: Any) -> Any:
        if self.custom_encoders:
            encoder = _get_encoder(obj, self.custom_encoders)
            if encoder is not None:
                return encoder(obj)

        if isinstance(obj, BSON_SCALAR_TYPES):
            return obj

        encoder = _get_encoder(obj, DEFAULT_CUSTOM_ENCODERS)
        if encoder is not None:
            return encoder(obj)

        if isinstance(obj, beanie.Document):
            return self._encode_document(obj)
        if IS_PYDANTIC_V2 and isinstance(obj, pydantic.RootModel):
            return self.encode(obj.root)
        if isinstance(obj, pydantic.BaseModel):
            items = self._iter_model_items(obj)
            return {key: self.encode(value) for key, value in items}
        if isinstance(obj, Mapping):
            return {
                key if isinstance(key, Enum) else str(key): self.encode(value)
                for key, value in obj.items()
            }
        if isinstance(obj, Iterable):
            return [self.encode(value) for value in obj]

        raise ValueError(f"Cannot encode {obj!r}")

    def _iter_model_items(
        self, obj: pydantic.BaseModel
    ) -> Iterable[Tuple[str, Any]]:
        keep_nulls = self.keep_nulls
        get_model_field = get_model_fields(obj).get
        for key, value in obj.__iter__():
            field_info = get_model_field(key)
            if field_info is not None:
                key = field_info.alias or key
            if not self._should_exclude_field(key, field_info) and (
                value is not None or keep_nulls
            ):
                yield key, value

    def _should_exclude_field(
        self, key: str, field_info: Optional[pydantic.fields.FieldInfo]
    ):
        exclude, include = (
            self.exclude,
            self.include,
        )

        if key in include:
            return False

        is_pydantic_excluded_field = (
            field_info is not None
            and (
                field_info.exclude
                if IS_PYDANTIC_V2
                else getattr(field_info.field_info, "exclude")
            )
            is True
        )
        return key in exclude or is_pydantic_excluded_field


def _get_encoder(
    obj: Any, custom_encoders: Mapping[type, SingleArgCallable]
) -> Optional[SingleArgCallable]:
    encoder = custom_encoders.get(type(obj))
    if encoder is not None:
        return encoder
    for cls, encoder in custom_encoders.items():
        if isinstance(obj, cls):
            return encoder
    return None



import pytest

from beanie.odm.operators.find.logical import And, Nor, Not, Or
from tests.odm.models import Sample


async def test_and():
    q = And(Sample.integer == 1)
    assert q == {"integer": 1}

    q = And(Sample.integer == 1, Sample.nested.integer > 3)
    assert q == {"$and": [{"integer": 1}, {"nested.integer": {"$gt": 3}}]}


async def test_not(preset_documents):
    q = Not(Sample.integer == 1)
    assert q == {"integer": {"$not": {"$eq": 1}}}

    docs = await Sample.find(q).to_list()
    assert len(docs) == 7

    with pytest.raises(AttributeError):
        q = Not(And(Sample.integer == 1, Sample.nested.integer > 3))
        await Sample.find(q).to_list()


async def test_nor():
    q = Nor(Sample.integer == 1)
    assert q == {"$nor": [{"integer": 1}]}

    q = Nor(Sample.integer == 1, Sample.nested.integer > 3)
    assert q == {"$nor": [{"integer": 1}, {"nested.integer": {"$gt": 3}}]}


async def test_or():
    q = Or(Sample.integer == 1)
    assert q == {"integer": 1}

    q = Or(Sample.integer == 1, Sample.nested.integer > 3)
    assert q == {"$or": [{"integer": 1}, {"nested.integer": {"$gt": 3}}]}



import pytest
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from tests.fastapi.app import app
from tests.fastapi.models import (
    DoorAPI,
    House,
    HouseAPI,
    Person,
    RoofAPI,
    WindowAPI,
)


@pytest.fixture(autouse=True)
async def api_client(clean_db):
    """api client fixture."""
    async with LifespanManager(app, startup_timeout=100, shutdown_timeout=100):
        server_name = "http://localhost"
        async with AsyncClient(
            transport=ASGITransport(app=app), base_url=server_name
        ) as ac:
            yield ac


@pytest.fixture(autouse=True)
async def clean_db(db):
    models = [House, Person, HouseAPI, WindowAPI, DoorAPI, RoofAPI]
    yield None

    for model in models:
        await model.get_pymongo_collection().drop()
        await model.get_pymongo_collection().drop_indexes()
